import hypothesis.strategies as st
import pytest
from hypothesis import given, settings

from vyper import compiler
from vyper.exceptions import (
    CallViolation,
    DuplicateImport,
    ImportCycle,
    StructureException,
    TypeMismatch,
)

# test modules which have no variables - "libraries"


def test_simple_library(get_contract, make_input_bundle, env):
    library_source = """
@internal
def foo() -> uint256:
    return block.number + 1
    """
    main = """
import library

@external
def bar() -> uint256:
    return library.foo() - 1
    """
    input_bundle = make_input_bundle({"library.vy": library_source})

    c = get_contract(main, input_bundle=input_bundle)

    assert c.bar() == env.block_number


# is this the best place for this?
def test_import_cycle(make_input_bundle):
    code_a = "import b\n"
    code_b = "import a\n"

    input_bundle = make_input_bundle({"a.vy": code_a, "b.vy": code_b})

    with pytest.raises(ImportCycle):
        compiler.compile_code(code_a, input_bundle=input_bundle)


# test we can have a function in the library with the same name as
# in the main contract
def test_library_function_same_name(get_contract, make_input_bundle):
    library = """
@internal
def foo() -> uint256:
    return 10
    """

    main = """
import library

@internal
def foo() -> uint256:
    return 100

@external
def self_foo() -> uint256:
    return self.foo()

@external
def library_foo() -> uint256:
    return library.foo()
    """

    input_bundle = make_input_bundle({"library.vy": library})

    c = get_contract(main, input_bundle=input_bundle)

    assert c.self_foo() == 100
    assert c.library_foo() == 10


def test_transitive_import(get_contract, make_input_bundle):
    a = """
@internal
def foo() -> uint256:
    return 1
    """
    b = """
import a

@internal
def bar() -> uint256:
    return a.foo() + 1
    """
    c = """
import b

@external
def baz() -> uint256:
    return b.bar() + 1
    """
    # more complicated call graph, with `a` imported twice.
    d = """
import b
import a

@external
def qux() -> uint256:
    s: uint256 = a.foo()
    return s + b.bar() + 1
    """
    input_bundle = make_input_bundle({"a.vy": a, "b.vy": b, "c.vy": c, "d.vy": d})

    contract = get_contract(c, input_bundle=input_bundle)
    assert contract.baz() == 3
    contract = get_contract(d, input_bundle=input_bundle)
    assert contract.qux() == 4


def test_cannot_call_library_external_functions(make_input_bundle):
    library_source = """
@external
def foo():
    pass
    """
    contract_source = """
import library

@external
def bar():
    library.foo()
    """
    input_bundle = make_input_bundle({"library.vy": library_source, "contract.vy": contract_source})
    with pytest.raises(CallViolation):
        compiler.compile_code(contract_source, input_bundle=input_bundle)


def test_library_external_functions_not_in_abi(get_contract, make_input_bundle):
    library_source = """
@external
def foo():
    pass
    """
    contract_source = """
import library

@external
def bar():
    pass
    """
    input_bundle = make_input_bundle({"library.vy": library_source, "contract.vy": contract_source})
    c = get_contract(contract_source, input_bundle=input_bundle)
    assert not hasattr(c, "foo")


def test_library_structs(get_contract, make_input_bundle):
    library_source = """
struct SomeStruct:
    x: uint256

@internal
def foo() -> SomeStruct:
    return SomeStruct(x=1)
    """
    contract_source = """
import library

@external
def bar(s: library.SomeStruct):
    pass

@external
def baz() -> library.SomeStruct:
    return library.SomeStruct(x=2)

@external
def qux() -> library.SomeStruct:
    return library.foo()
    """
    input_bundle = make_input_bundle({"library.vy": library_source, "contract.vy": contract_source})
    c = get_contract(contract_source, input_bundle=input_bundle)

    assert c.bar((1,)) is None

    assert c.baz() == (2,)
    assert c.qux() == (1,)


# test calls to library functions in statement position
def test_library_statement_calls(get_contract, make_input_bundle, tx_failed):
    library_source = """
from ethereum.ercs import IERC20
@internal
def check_adds_to_ten(x: uint256, y: uint256):
    assert x + y == 10
    """
    contract_source = """
import library

counter: public(uint256)

@external
def foo(x: uint256):
    library.check_adds_to_ten(3, x)
    self.counter = x
    """
    input_bundle = make_input_bundle({"library.vy": library_source, "contract.vy": contract_source})

    c = get_contract(contract_source, input_bundle=input_bundle)

    c.foo(7)

    assert c.counter() == 7

    with tx_failed():
        c.foo(8)


def test_library_is_typechecked(make_input_bundle):
    library_source = """
@internal
def foo():
    asdlkfjasdflkajsdf
    """
    contract_source = """
import library
    """

    input_bundle = make_input_bundle({"library.vy": library_source, "contract.vy": contract_source})
    with pytest.raises(StructureException):
        compiler.compile_code(contract_source, input_bundle=input_bundle)


def test_library_is_typechecked2(make_input_bundle):
    # check that we typecheck against imported function signatures
    library_source = """
@internal
def foo() -> uint256:
    return 1
    """
    contract_source = """
import library

@external
def foo() -> bytes32:
    return library.foo()
    """

    input_bundle = make_input_bundle({"library.vy": library_source, "contract.vy": contract_source})
    with pytest.raises(TypeMismatch):
        compiler.compile_code(contract_source, input_bundle=input_bundle)


def test_reject_duplicate_imports(make_input_bundle):
    library_source = """
    """

    contract_source = """
import library
import library as library2
    """
    input_bundle = make_input_bundle({"library.vy": library_source, "contract.vy": contract_source})
    with pytest.raises(DuplicateImport):
        compiler.compile_code(contract_source, input_bundle=input_bundle)


def test_nested_module_access(get_contract, make_input_bundle):
    lib1 = """
import lib2

@internal
def lib2_foo() -> uint256:
    return lib2.foo()
    """
    lib2 = """
@internal
def foo() -> uint256:
    return 1337
    """

    main = """
import lib1
import lib2

@external
def lib1_foo() -> uint256:
    return lib1.lib2_foo()

@external
def lib2_foo() -> uint256:
    return lib1.lib2.foo()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})
    c = get_contract(main, input_bundle=input_bundle)

    assert c.lib1_foo() == c.lib2_foo() == 1337


_int_127 = st.integers(min_value=0, max_value=127)
_bytes_128 = st.binary(min_size=0, max_size=128)


def test_slice_builtin(get_contract, make_input_bundle):
    lib = """
@internal
def slice_input(x: Bytes[128], start: uint256, length: uint256) -> Bytes[128]:
    return slice(x, start, length)
    """

    main = """
import lib
@external
def lib_slice_input(x: Bytes[128], start: uint256, length: uint256) -> Bytes[128]:
    return lib.slice_input(x, start, length)

@external
def slice_input(x: Bytes[128], start: uint256, length: uint256) -> Bytes[128]:
    return slice(x, start, length)
    """
    input_bundle = make_input_bundle({"lib.vy": lib})
    c = get_contract(main, input_bundle=input_bundle)

    # use an inner test so that we can cache the result of get_contract()
    @given(start=_int_127, length=_int_127, bytesdata=_bytes_128)
    @settings(max_examples=100)
    def _test(bytesdata, start, length):
        # surjectively map start into allowable range
        if start > len(bytesdata):
            start = start % (len(bytesdata) or 1)
        # surjectively map length into allowable range
        if length > (len(bytesdata) - start):
            length = length % ((len(bytesdata) - start) or 1)
        main_result = c.slice_input(bytesdata, start, length)
        library_result = c.lib_slice_input(bytesdata, start, length)
        assert main_result == library_result == bytesdata[start : start + length]

    _test()



import json
from pathlib import Path, PurePath

import pytest

from tests.utils import working_directory
from vyper.compiler.input_bundle import ABIInput, FileInput, FilesystemInputBundle, JSONInputBundle


# FilesystemInputBundle which uses same search path as make_file
@pytest.fixture
def input_bundle(tmp_path):
    return FilesystemInputBundle([tmp_path])


def test_load_file(make_file, input_bundle):
    filepath = make_file("foo.vy", "contents")

    file = input_bundle.load_file(Path("foo.vy"))

    assert isinstance(file, FileInput)
    assert file == FileInput(0, Path("foo.vy"), filepath, "contents")


def test_search_path_context_manager(make_file, tmp_path):
    ib = FilesystemInputBundle([])

    filepath = make_file("foo.vy", "contents")

    with pytest.raises(FileNotFoundError):
        # no search path given
        ib.load_file(Path("foo.vy"))

    with ib.search_path(tmp_path):
        file = ib.load_file(Path("foo.vy"))

    assert isinstance(file, FileInput)
    assert file == FileInput(0, Path("foo.vy"), filepath, "contents")


def test_search_path_precedence(make_file, tmp_path, tmp_path_factory, input_bundle):
    # test search path precedence.
    # most recent search path is the highest precedence
    tmpdir = tmp_path_factory.mktemp("some_directory")
    tmpdir2 = tmp_path_factory.mktemp("some_other_directory")

    filepaths = []
    for i, directory in enumerate([tmp_path, tmpdir, tmpdir2]):
        path = directory / "foo.vy"
        with path.open("w") as f:
            f.write(f"contents {i}")
        filepaths.append(path)

    ib = FilesystemInputBundle([tmp_path, tmpdir, tmpdir2])

    file = ib.load_file("foo.vy")

    assert isinstance(file, FileInput)
    assert file == FileInput(0, Path("foo.vy"), filepaths[2], "contents 2")

    with ib.search_path(tmpdir):
        file = ib.load_file("foo.vy")

        assert isinstance(file, FileInput)
        assert file == FileInput(1, PurePath("foo.vy"), filepaths[1], "contents 1")


# special rules for handling json files
def test_load_abi(make_file, input_bundle, tmp_path):
    contents = json.dumps("some string")

    path = make_file("foo.json", contents)

    file = input_bundle.load_file("foo.json")
    assert isinstance(file, ABIInput)
    assert file == ABIInput(0, PurePath("foo.json"), path, contents, "some string")

    # suffix doesn't matter
    path = make_file("foo.txt", contents)
    file = input_bundle.load_file("foo.txt")
    assert isinstance(file, ABIInput)
    assert file == ABIInput(1, PurePath("foo.txt"), path, contents, "some string")


# check that unique paths give unique source ids
def test_source_id_file_input(make_file, input_bundle, tmp_path):
    foopath = make_file("foo.vy", "contents")
    barpath = make_file("bar.vy", "contents 2")

    file = input_bundle.load_file("foo.vy")
    assert file.source_id == 0
    assert file == FileInput(0, PurePath("foo.vy"), foopath, "contents")

    file2 = input_bundle.load_file("bar.vy")
    # source id increments
    assert file2.source_id == 1
    assert file2 == FileInput(1, PurePath("bar.vy"), barpath, "contents 2")

    file3 = input_bundle.load_file("foo.vy")
    assert file3.source_id == 0
    assert file3 == FileInput(0, PurePath("foo.vy"), foopath, "contents")

    # test source id is stable across different search paths
    with working_directory(tmp_path):
        with input_bundle.search_path(Path(".")):
            file4 = input_bundle.load_file("foo.vy")
            assert file4.source_id == 0
            assert file4 == FileInput(0, PurePath("foo.vy"), foopath, "contents")

    # test source id is stable even when requested filename is different
    with working_directory(tmp_path.parent):
        with input_bundle.search_path(Path(".")):
            file5 = input_bundle.load_file(Path(tmp_path.stem) / "foo.vy")
            assert file5.source_id == 0
            assert file5 == FileInput(0, Path(tmp_path.stem) / "foo.vy", foopath, "contents")


# check that unique paths give unique source ids
def test_source_id_json_input(make_file, input_bundle, tmp_path):
    contents = json.dumps("some string")
    contents2 = json.dumps(["some list"])

    foopath = make_file("foo.json", contents)

    barpath = make_file("bar.json", contents2)

    file = input_bundle.load_file("foo.json")
    assert isinstance(file, ABIInput)
    assert file == ABIInput(0, PurePath("foo.json"), foopath, contents, "some string")

    file2 = input_bundle.load_file("bar.json")
    assert isinstance(file2, ABIInput)
    assert file2 == ABIInput(1, PurePath("bar.json"), barpath, contents2, ["some list"])

    file3 = input_bundle.load_file("foo.json")
    assert file3.source_id == 0
    assert file3 == ABIInput(0, PurePath("foo.json"), foopath, contents, "some string")

    # test source id is stable across different search paths
    with working_directory(tmp_path):
        with input_bundle.search_path(Path(".")):
            file4 = input_bundle.load_file("foo.json")
            assert file4.source_id == 0
            assert file4 == ABIInput(0, PurePath("foo.json"), foopath, contents, "some string")

    # test source id is stable even when requested filename is different
    with working_directory(tmp_path.parent):
        with input_bundle.search_path(Path(".")):
            file5 = input_bundle.load_file(Path(tmp_path.stem) / "foo.json")
            assert file5.source_id == 0
            assert file5 == ABIInput(
                0, Path(tmp_path.stem) / "foo.json", foopath, contents, "some string"
            )


# test some pathological case where the file changes underneath
def test_mutating_file_source_id(make_file, input_bundle, tmp_path):
    foopath = make_file("foo.vy", "contents")

    file = input_bundle.load_file("foo.vy")
    assert file.source_id == 0
    assert file == FileInput(0, PurePath("foo.vy"), foopath, "contents")

    foopath = make_file("foo.vy", "new contents")

    file = input_bundle.load_file("foo.vy")
    # source id hasn't changed, even though contents have
    assert file.source_id == 0
    assert file == FileInput(0, PurePath("foo.vy"), foopath, "new contents")


# test the os.normpath behavior of symlink
# (slightly pathological, for illustration's sake)
def test_load_file_symlink(make_file, input_bundle, tmp_path, tmp_path_factory):
    dir1 = tmp_path / "first"
    dir2 = tmp_path / "second"
    symlink = tmp_path / "symlink"

    dir1.mkdir()
    dir2.mkdir()
    symlink.symlink_to(dir2, target_is_directory=True)

    outer_path = tmp_path / "foo.vy"
    with outer_path.open("w") as f:
        f.write("contents of the outer directory")

    inner_path = dir1 / "foo.vy"
    with inner_path.open("w") as f:
        f.write("contents of the inner directory")

    # symlink rules would be:
    # base/symlink/../foo.vy =>
    # base/first/second/../foo.vy =>
    # base/first/foo.vy
    # normpath would be base/symlink/../foo.vy =>
    # base/foo.vy
    to_load = symlink / ".." / "foo.vy"
    file = input_bundle.load_file(to_load)

    assert file == FileInput(0, to_load, outer_path.resolve(), "contents of the outer directory")


def test_json_input_bundle_basic():
    files = {PurePath("foo.vy"): {"content": "some text"}}
    input_bundle = JSONInputBundle(files, [PurePath(".")])

    file = input_bundle.load_file(PurePath("foo.vy"))
    assert file == FileInput(0, PurePath("foo.vy"), PurePath("foo.vy"), "some text")


def test_json_input_bundle_normpath():
    contents = "some text"
    files = {PurePath("foo/../bar.vy"): {"content": contents}}
    input_bundle = JSONInputBundle(files, [PurePath(".")])

    barpath = PurePath("bar.vy")

    expected = FileInput(0, barpath, barpath, contents)

    file = input_bundle.load_file(PurePath("bar.vy"))
    assert file == expected

    file = input_bundle.load_file(PurePath("baz/../bar.vy"))
    assert file == FileInput(0, PurePath("baz/../bar.vy"), barpath, contents)

    file = input_bundle.load_file(PurePath("./bar.vy"))
    assert file == FileInput(0, PurePath("./bar.vy"), barpath, contents)

    with input_bundle.search_path(PurePath("foo")):
        file = input_bundle.load_file(PurePath("../bar.vy"))
        assert file == FileInput(0, PurePath("../bar.vy"), barpath, contents)


def test_json_input_abi():
    some_abi = ["some abi"]
    some_abi_str = json.dumps(some_abi)
    foopath = PurePath("foo.json")
    barpath = PurePath("bar.txt")
    files = {foopath: {"abi": some_abi}, barpath: {"content": some_abi_str}}
    input_bundle = JSONInputBundle(files, [PurePath(".")])

    file = input_bundle.load_file(foopath)
    abi_contents = json.dumps({"abi": some_abi})
    assert file == ABIInput(0, foopath, foopath, abi_contents, some_abi)

    file = input_bundle.load_file(barpath)
    assert file == ABIInput(1, barpath, barpath, some_abi_str, some_abi)



def test_constant_address_balance(env, get_contract):
    code = """
a: constant(address) = 0x776Ba14735FF84789320718cf0aa43e91F7A8Ce1

@external
def foo() -> uint256:
    x: uint256 = a.balance
    return x
    """
    address = "0x776Ba14735FF84789320718cf0aa43e91F7A8Ce1"

    c = get_contract(code)

    assert c.foo() == 0

    env.set_balance(env.deployer, 1337)
    env.message_call(address, value=1337)

    assert c.foo() == 1337



from tests.utils import decimal_to_int


def test_call_to_self_struct(env, get_contract):
    code = """
struct MyStruct:
    e1: decimal
    e2: uint256

@internal
@view
def get_my_struct(_e1: decimal, _e2: uint256) -> MyStruct:
    return MyStruct(e1=_e1, e2=_e2)

@external
@view
def wrap_get_my_struct_WORKING(_e1: decimal) -> MyStruct:
    testing: MyStruct = self.get_my_struct(_e1, block.timestamp)
    return testing

@external
@view
def wrap_get_my_struct_BROKEN(_e1: decimal) -> MyStruct:
    return self.get_my_struct(_e1, block.timestamp)
    """
    c = get_contract(code)
    assert c.wrap_get_my_struct_WORKING(decimal_to_int("0.1")) == (
        decimal_to_int("0.1"),
        env.timestamp,
    )
    assert c.wrap_get_my_struct_BROKEN(decimal_to_int("0.1")) == (
        decimal_to_int("0.1"),
        env.timestamp,
    )


def test_call_to_self_struct_2(get_contract):
    code = """
struct MyStruct:
    e1: decimal

@internal
@view
def get_my_struct(_e1: decimal) -> MyStruct:
    return MyStruct(e1=_e1)

@external
@view
def wrap_get_my_struct_WORKING(_e1: decimal) -> MyStruct:
    testing: MyStruct = self.get_my_struct(_e1)
    return testing

@external
@view
def wrap_get_my_struct_BROKEN(_e1: decimal) -> MyStruct:
    return self.get_my_struct(_e1)
    """
    c = get_contract(code)
    assert c.wrap_get_my_struct_WORKING(decimal_to_int("0.1")) == (decimal_to_int("0.1"),)
    assert c.wrap_get_my_struct_BROKEN(decimal_to_int("0.1")) == (decimal_to_int("0.1"),)



import pytest

from vyper.exceptions import TypeMismatch


def test_basic_bytes_keys(env, get_contract):
    code = """
mapped_bytes: HashMap[Bytes[5], int128]

@external
def set(k: Bytes[5], v: int128):
    self.mapped_bytes[k] = v

@external
def get(k: Bytes[5]) -> int128:
    return self.mapped_bytes[k]
    """

    c = get_contract(code)

    c.set(b"test", 54321)

    assert c.get(b"test") == 54321


def test_basic_bytes_literal_key(get_contract):
    code = """
mapped_bytes: HashMap[Bytes[5], int128]

@external
def set(v: int128):
    self.mapped_bytes[b"test"] = v

@external
def get(k: Bytes[5]) -> int128:
    return self.mapped_bytes[k]
    """

    c = get_contract(code)

    c.set(54321)

    assert c.get(b"test") == 54321


def test_basic_long_bytes_as_keys(get_contract):
    code = """
mapped_bytes: HashMap[Bytes[34], int128]

@external
def set(k: Bytes[34], v: int128):
    self.mapped_bytes[k] = v

@external
def get(k: Bytes[34]) -> int128:
    return self.mapped_bytes[k]
    """

    c = get_contract(code)

    c.set(b"a" * 34, 6789, gas=10**6)

    assert c.get(b"a" * 34) == 6789


def test_mismatched_byte_length(get_contract):
    code = """
mapped_bytes: HashMap[Bytes[34], int128]

@external
def set(k: Bytes[35], v: int128):
    self.mapped_bytes[k] = v
    """

    with pytest.raises(TypeMismatch):
        get_contract(code)


def test_extended_bytes_key_from_storage(get_contract):
    code = """
a: HashMap[Bytes[100000], int128]

@deploy
def __init__():
    self.a[b"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"] = 1069

@external
def get_it1() -> int128:
    key: Bytes[100000] = b"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
    return self.a[key]

@external
def get_it2() -> int128:
    return self.a[b"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"]

@external
def get_it3(key: Bytes[100000]) -> int128:
    return self.a[key]
    """

    c = get_contract(code)

    assert c.get_it2() == 1069
    assert c.get_it2() == 1069
    assert c.get_it3(b"a" * 33) == 1069
    assert c.get_it3(b"test") == 0


def test_struct_bytes_key_memory(get_contract):
    code = """
struct Foo:
    one: Bytes[5]
    two: Bytes[100]

a: HashMap[Bytes[100000], int128]

@deploy
def __init__():
    self.a[b"hello"] = 1069
    self.a[b"potato"] = 31337

@external
def get_one() -> int128:
    b: Foo = Foo(one=b"hello", two=b"potato")
    return self.a[b.one]

@external
def get_two() -> int128:
    b: Foo = Foo(one=b"hello", two=b"potato")
    return self.a[b.two]
"""

    c = get_contract(code)

    assert c.get_one() == 1069
    assert c.get_two() == 31337


def test_struct_bytes_key_storage(get_contract):
    code = """
struct Foo:
    one: Bytes[5]
    two: Bytes[100]

a: HashMap[Bytes[100000], int128]
b: Foo

@deploy
def __init__():
    self.a[b"hello"] = 1069
    self.a[b"potato"] = 31337
    self.b = Foo(one=b"hello", two=b"potato")

@external
def get_one() -> int128:
    return self.a[self.b.one]

@external
def get_two() -> int128:
    return self.a[self.b.two]
"""

    c = get_contract(code)

    assert c.get_one() == 1069
    assert c.get_two() == 31337


def test_bytes_key_storage(get_contract):
    code = """

a: HashMap[Bytes[100000], int128]
b: Bytes[5]

@deploy
def __init__():
    self.a[b"hello"] = 1069
    self.b = b"hello"

@external
def get_storage() -> int128:
    return self.a[self.b]
"""

    c = get_contract(code)

    assert c.get_storage() == 1069


def test_bytes_key_calldata(get_contract):
    code = """

a: HashMap[Bytes[100000], int128]


@deploy
def __init__():
    self.a[b"hello"] = 1069

@external
def get_calldata(b: Bytes[5]) -> int128:
    return self.a[b]
"""

    c = get_contract(code)

    assert c.get_calldata(b"hello") == 1069


def test_struct_bytes_hashmap_as_key_in_other_hashmap(get_contract):
    code = """
struct Thing:
    name: Bytes[64]

bar: public(HashMap[uint256, Thing])
foo: public(HashMap[Bytes[64], uint256])

@deploy
def __init__():
    self.foo[b"hello"] = 31337
    self.bar[12] = Thing(name=b"hello")

@external
def do_the_thing(_index: uint256) -> uint256:
    return self.foo[self.bar[_index].name]
    """

    c = get_contract(code)

    assert c.do_the_thing(12) == 31337



import pytest

from vyper.compiler import compile_code
from vyper.compiler.output import build_abi_output
from vyper.compiler.phases import CompilerData

source_codes = [
    """
x: int128

@deploy
def __init__():
    self.x = 1
    """,
    """
x: int128

@deploy
def __init__():
    pass
    """,
]


@pytest.mark.parametrize("source_code", source_codes)
def test_only_init_function(source_code):
    empty_sig = [
        {"outputs": [], "inputs": [], "stateMutability": "nonpayable", "type": "constructor"}
    ]

    data = CompilerData(source_code)
    assert build_abi_output(data) == empty_sig


def test_default_abi():
    default_code = """
@payable
@external
def __default__():
    pass
    """

    data = CompilerData(default_code)
    assert build_abi_output(data) == [{"stateMutability": "payable", "type": "fallback"}]


def test_method_identifiers():
    code = """
x: public(int128)

@external
def foo(y: uint256) -> Bytes[100]:
    return b"hello"
    """

    out = compile_code(code, output_formats=["method_identifiers"])

    assert out["method_identifiers"] == {"foo(uint256)": "0x2fbebd38", "x()": "0xc55699c"}


def test_decimal_abi():
    code = """
x: public(decimal)

@external
@view
def foo(s: decimal) -> decimal:
    return s
    """

    out = compile_code(code, output_formats=["method_identifiers", "abi"])

    assert out["method_identifiers"] == {"foo(int168)": "0x929893b6", "x()": "0xc55699c"}

    expected_abi = [
        {
            "stateMutability": "view",
            "type": "function",
            "name": "foo",
            "inputs": [{"name": "s", "type": "int168", "internalType": "decimal"}],
            "outputs": [{"name": "", "type": "int168", "internalType": "decimal"}],
        },
        {
            "stateMutability": "view",
            "type": "function",
            "name": "x",
            "inputs": [],
            "outputs": [{"name": "", "type": "int168", "internalType": "decimal"}],
        },
    ]
    assert out["abi"] == expected_abi


def test_struct_abi():
    code = """
struct MyStruct:
    a: address
    b: uint256

@external
@view
def foo(s: MyStruct) -> MyStruct:
    return s
    """

    data = CompilerData(code)
    abi = build_abi_output(data)
    func_abi = abi[0]

    assert func_abi["name"] == "foo"

    expected_output = [
        {
            "type": "tuple",
            "name": "",
            "components": [{"type": "address", "name": "a"}, {"type": "uint256", "name": "b"}],
        }
    ]

    assert func_abi["outputs"] == expected_output

    expected_input = {
        "type": "tuple",
        "name": "s",
        "components": [{"type": "address", "name": "a"}, {"type": "uint256", "name": "b"}],
    }

    assert func_abi["inputs"][0] == expected_input


@pytest.mark.parametrize(
    "type,abi_type", [("DynArray[NestedStruct, 2]", "tuple[]"), ("NestedStruct[2]", "tuple[2]")]
)
def test_nested_struct(type, abi_type):
    code = f"""
struct MyStruct:
    a: address
    b: bytes32

struct NestedStruct:
    t: MyStruct
    foo: uint256

@view
@external
def getStructList() -> {type}:
    return [
        NestedStruct(t=MyStruct(a=msg.sender, b=block.prevhash), foo=1),
        NestedStruct(t=MyStruct(a=msg.sender, b=block.prevhash), foo=2)
    ]
    """

    out = compile_code(code, output_formats=["abi"])

    assert out["abi"] == [
        {
            "inputs": [],
            "name": "getStructList",
            "outputs": [
                {
                    "components": [
                        {
                            "components": [
                                {"name": "a", "type": "address"},
                                {"name": "b", "type": "bytes32"},
                            ],
                            "name": "t",
                            "type": "tuple",
                        },
                        {"name": "foo", "type": "uint256"},
                    ],
                    "name": "",
                    "type": f"{abi_type}",
                }
            ],
            "stateMutability": "view",
            "type": "function",
        }
    ]


@pytest.mark.parametrize(
    "type,abi_type", [("DynArray[DynArray[Foo, 2], 2]", "tuple[][]"), ("Foo[2][2]", "tuple[2][2]")]
)
def test_2d_list_of_struct(type, abi_type):
    code = f"""
struct Foo:
    a: uint256
    b: uint256

@view
@external
def bar(x: {type}):
    pass
    """

    out = compile_code(code, output_formats=["abi"])

    assert out["abi"] == [
        {
            "inputs": [
                {
                    "components": [
                        {"name": "a", "type": "uint256"},
                        {"name": "b", "type": "uint256"},
                    ],
                    "name": "x",
                    "type": f"{abi_type}",
                }
            ],
            "name": "bar",
            "outputs": [],
            "stateMutability": "view",
            "type": "function",
        }
    ]


def test_exports_abi(make_input_bundle):
    lib1 = """
@external
def foo():
    pass

@external
def bar():
    pass
    """

    main = """
import lib1

initializes: lib1

exports: lib1.foo
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    out = compile_code(main, input_bundle=input_bundle, output_formats=["abi"])

    # just for clarity -- check bar() is not in the output
    for fn in out["abi"]:
        assert fn["name"] != "bar"

    expected = [
        {
            "inputs": [],
            "name": "foo",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        }
    ]

    assert out["abi"] == expected


def test_exports_variable(make_input_bundle):
    lib1 = """
@external
def foo():
    pass

private_storage_variable: uint256
private_immutable_variable: immutable(uint256)
private_constant_variable: constant(uint256) = 3

public_storage_variable: public(uint256)
public_immutable_variable: public(immutable(uint256))
public_constant_variable: public(constant(uint256)) = 10

@deploy
def __init__(a: uint256, b: uint256):
    public_immutable_variable = a
    private_immutable_variable = b
    """

    main = """
import lib1

initializes: lib1

exports: (
    lib1.foo,
    lib1.public_storage_variable,
    lib1.public_immutable_variable,
    lib1.public_constant_variable,
)

@deploy
def __init__():
    lib1.__init__(5, 6)
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    out = compile_code(main, input_bundle=input_bundle, output_formats=["abi"])
    expected = [
        {
            "inputs": [],
            "name": "foo",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "public_storage_variable",
            "outputs": [{"name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "public_immutable_variable",
            "outputs": [{"name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [],
            "name": "public_constant_variable",
            "outputs": [{"name": "", "type": "uint256"}],
            "stateMutability": "view",
            "type": "function",
        },
        {"inputs": [], "outputs": [], "stateMutability": "nonpayable", "type": "constructor"},
    ]

    assert out["abi"] == expected


def test_event_export_from_init(make_input_bundle):
    # test that events get exported when used in init functions
    lib1 = """
event MyEvent:
    pass

@deploy
def __init__():
    log MyEvent()
    """
    main = """
import lib1

initializes: lib1

@deploy
def __init__():
    lib1.__init__()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    out = compile_code(main, input_bundle=input_bundle, output_formats=["abi"])
    expected = {
        "abi": [
            {"anonymous": False, "inputs": [], "name": "MyEvent", "type": "event"},
            {"inputs": [], "outputs": [], "stateMutability": "nonpayable", "type": "constructor"},
        ]
    }

    assert out == expected


def test_event_export_from_function_export(make_input_bundle):
    # test events used in exported functions are exported
    lib1 = """
event MyEvent:
    pass

@external
def foo():
    log MyEvent()
    """
    main = """
import lib1

initializes: lib1

exports: lib1.foo
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    out = compile_code(main, input_bundle=input_bundle, output_formats=["abi"])
    expected = {
        "abi": [
            {"anonymous": False, "inputs": [], "name": "MyEvent", "type": "event"},
            {
                "name": "foo",
                "inputs": [],
                "outputs": [],
                "stateMutability": "nonpayable",
                "type": "function",
            },
        ]
    }

    assert out == expected


def test_event_export_unused_function(make_input_bundle):
    # test events in unused functions are not exported
    lib1 = """
event MyEvent:
    pass

@internal
def foo():
    log MyEvent()
    """
    main = """
import lib1
initializes: lib1

# not exported/reachable from selector table
@internal
def do_foo():
    lib1.foo()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    out = compile_code(main, input_bundle=input_bundle, output_formats=["abi"])
    expected = {"abi": []}

    assert out == expected


def test_event_export_unused_module(make_input_bundle):
    # test events are exported from functions which are used, even
    # if the module is not marked `uses:`.
    lib1 = """
event MyEvent:
    pass

@internal
def foo():
    log MyEvent()
    """
    main = """
import lib1

@external
def bar():
    lib1.foo()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    out = compile_code(main, input_bundle=input_bundle, output_formats=["abi"])
    expected = {
        "abi": [
            {"anonymous": False, "inputs": [], "name": "MyEvent", "type": "event"},
            {
                "inputs": [],
                "name": "bar",
                "outputs": [],
                "stateMutability": "nonpayable",
                "type": "function",
            },
        ]
    }

    assert out == expected


def test_event_no_export_implements(make_input_bundle):
    # test events are not exported even if they are in implemented interface
    ifoo = """
event MyEvent:
    pass
    """
    main = """
import ifoo

implements: ifoo
    """
    input_bundle = make_input_bundle({"ifoo.vyi": ifoo})
    out = compile_code(main, input_bundle=input_bundle, output_formats=["abi"])
    expected = {"abi": []}

    assert out == expected


def test_event_export_interface(make_input_bundle):
    # test events from interfaces get exported
    ifoo = """
event MyEvent:
    pass

@external
def foo():
    ...
    """
    main = """
import ifoo

@external
def bar():
    log ifoo.MyEvent()
    """
    input_bundle = make_input_bundle({"ifoo.vyi": ifoo})
    out = compile_code(main, input_bundle=input_bundle, output_formats=["abi"])
    expected = {
        "abi": [
            {"anonymous": False, "inputs": [], "name": "MyEvent", "type": "event"},
            {
                "inputs": [],
                "name": "bar",
                "outputs": [],
                "stateMutability": "nonpayable",
                "type": "function",
            },
        ]
    }
    assert out == expected


def test_event_export_interface_no_use(make_input_bundle):
    # test events from interfaces don't get exported unless used
    ifoo = """
event MyEvent:
    pass

@external
def foo():
    ...
    """
    main = """
import ifoo

@external
def bar():
    extcall ifoo(msg.sender).foo()
    """
    input_bundle = make_input_bundle({"ifoo.vyi": ifoo})
    out = compile_code(main, input_bundle=input_bundle, output_formats=["abi"])
    expected = {
        "abi": [
            {
                "inputs": [],
                "name": "bar",
                "outputs": [],
                "stateMutability": "nonpayable",
                "type": "function",
            }
        ]
    }

    assert out == expected


def test_event_export_nested_export_chain(make_input_bundle):
    # test exporting an event from a nested used module
    lib1 = """
event MyEvent:
    pass

@external
def foo():
    log MyEvent()
    """
    lib2 = """
import lib1
exports: lib1.foo
    """
    main = """
import lib2
exports: lib2.lib1.foo
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})
    out = compile_code(main, input_bundle=input_bundle, output_formats=["abi"])
    expected = {
        "abi": [
            {"anonymous": False, "inputs": [], "name": "MyEvent", "type": "event"},
            {
                "name": "foo",
                "inputs": [],
                "outputs": [],
                "stateMutability": "nonpayable",
                "type": "function",
            },
        ]
    }

    assert out == expected


def test_event_export_nested_internal(make_input_bundle):
    # test events are exported from nested internal calls across modules
    lib1 = """
event MyEvent:
    pass

@internal
def foo():
    log MyEvent()
    """
    lib2 = """
import lib1

@internal
def bar():
    lib1.foo()
    """
    main = """
import lib2  # no uses

@external
def baz():
    lib2.bar()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})
    out = compile_code(main, input_bundle=input_bundle, output_formats=["abi"])
    expected = {
        "abi": [
            {"anonymous": False, "inputs": [], "name": "MyEvent", "type": "event"},
            {
                "name": "baz",
                "inputs": [],
                "outputs": [],
                "stateMutability": "nonpayable",
                "type": "function",
            },
        ]
    }

    assert out == expected


def test_event_export_nested_no_uses(make_input_bundle):
    # event is not exported when it's not used
    lib1 = """
event MyEvent:
    pass

counter: uint256

@internal
def foo():
    log MyEvent()

@internal
def update_counter():
    self.counter += 1
    """
    lib2 = """
import lib1
uses: lib1

@internal
def use_lib1():
    lib1.update_counter()
    """
    main = """
import lib1
import lib2

initializes: lib1
initializes: lib2[lib1 := lib1]

@external
def foo():
    lib2.use_lib1()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})
    out = compile_code(main, input_bundle=input_bundle, output_formats=["abi"])
    expected = {
        "abi": [
            {
                "name": "foo",
                "inputs": [],
                "outputs": [],
                "stateMutability": "nonpayable",
                "type": "function",
            }
        ]
    }

    assert out == expected



#!/usr/bin/env python3
import argparse
import json
import os
import sys
import warnings
from pathlib import Path
from typing import Any, Iterable, Iterator, Optional, Set, TypeVar

import vyper
import vyper.codegen.ir_node as ir_node
import vyper.evm.opcodes as evm
from vyper.cli import vyper_json
from vyper.cli.compile_archive import NotZipInput, compile_from_zip
from vyper.compiler.input_bundle import FileInput, FilesystemInputBundle
from vyper.compiler.settings import VYPER_TRACEBACK_LIMIT, OptimizationLevel, Settings
from vyper.typing import ContractPath, OutputFormats

T = TypeVar("T")

format_options_help = """Format to print, one or more of:
bytecode (default) - Deployable bytecode
bytecode_runtime   - Bytecode at runtime
blueprint_bytecode - Deployment bytecode for an ERC-5202 compatible blueprint
abi                - ABI in JSON format
abi_python         - ABI in python format
source_map         - Vyper source map of deployable bytecode
source_map_runtime - Vyper source map of runtime bytecode
method_identifiers - Dictionary of method signature to method identifier
userdoc            - Natspec user documentation
devdoc             - Natspec developer documentation
metadata           - Contract metadata (intended for use by tooling developers)
combined_json      - All of the above format options combined as single JSON output
layout             - Storage layout of a Vyper contract
ast                - AST (not yet annotated) in JSON format
annotated_ast      - Annotated AST in JSON format
interface          - Vyper interface of a contract
external_interface - External interface of a contract, used for outside contract calls
opcodes            - List of opcodes as a string
opcodes_runtime    - List of runtime opcodes as a string
ir                 - Intermediate representation in list format
ir_json            - Intermediate representation in JSON format
ir_runtime         - Intermediate representation of runtime bytecode in list format
asm                - Output the EVM assembly of the deployable bytecode
archive            - Output the build as an archive file
solc_json          - Output the build in solc json format
"""

combined_json_outputs = [
    "bytecode",
    "bytecode_runtime",
    "blueprint_bytecode",
    "abi",
    "layout",
    "source_map",
    "source_map_runtime",
    "method_identifiers",
    "userdoc",
    "devdoc",
]


def _parse_cli_args():
    return _parse_args(sys.argv[1:])


def _cli_helper(f, output_formats, compiled):
    if output_formats == ("combined_json",):
        compiled = {str(path): v for (path, v) in compiled.items()}
        print(json.dumps(compiled), file=f)
        return

    if output_formats == ("archive",):
        for contract_data in compiled.values():
            assert list(contract_data.keys()) == ["archive"]
            out = contract_data["archive"]
            if f.isatty() and isinstance(out, bytes):
                raise RuntimeError(
                    "won't write raw bytes to a tty! (if you want to base64"
                    " encode the archive, you can try `-f archive` in"
                    " conjunction with `--base64`)"
                )
            else:
                f.write(out)
        return

    for contract_data in compiled.values():
        for data in contract_data.values():
            if isinstance(data, (list, dict)):
                print(json.dumps(data), file=f)
            else:
                print(data, file=f)


def _parse_args(argv):
    warnings.simplefilter("always")

    if "--standard-json" in argv:
        argv.remove("--standard-json")
        vyper_json._parse_args(argv)
        return

    parser = argparse.ArgumentParser(
        description="Pythonic Smart Contract Language for the EVM",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument("input_files", help="Vyper sourcecode to compile", nargs="+")
    parser.add_argument("--version", action="version", version=vyper.__long_version__)
    parser.add_argument(
        "--show-gas-estimates",
        help="Show gas estimates in abi and ir output mode.",
        action="store_true",
    )
    parser.add_argument("-f", help=format_options_help, default="bytecode", dest="format")
    parser.add_argument(
        "--storage-layout-file",
        help="Override storage slots provided by compiler",
        dest="storage_layout",
        nargs="+",
    )
    parser.add_argument(
        "--evm-version",
        help=f"Select desired EVM version (default {evm.DEFAULT_EVM_VERSION}). "
        "note: cancun support is EXPERIMENTAL",
        choices=list(evm.EVM_VERSIONS),
        dest="evm_version",
    )
    parser.add_argument("--no-optimize", help="Do not optimize", action="store_true")
    parser.add_argument(
        "--base64",
        help="Base64 encode the output (only valid in conjunction with `-f archive`",
        action="store_true",
    )
    parser.add_argument(
        "-O",
        "--optimize",
        help="Optimization flag (defaults to 'gas')",
        choices=["gas", "codesize", "none"],
    )
    parser.add_argument("--debug", help="Compile in debug mode", action="store_true")
    parser.add_argument(
        "--no-bytecode-metadata", help="Do not add metadata to bytecode", action="store_true"
    )
    parser.add_argument(
        "--traceback-limit",
        help="Set the traceback limit for error messages reported by the compiler",
        type=int,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="Turn on compiler verbose output. "
        "Currently an alias for --traceback-limit but "
        "may add more information in the future",
        action="store_true",
    )
    parser.add_argument(
        "--standard-json",
        help="Switch to standard JSON mode. Use `--standard-json -h` for available options.",
        action="store_true",
    )
    parser.add_argument(
        "--hex-ir", help="Represent integers as hex values in the IR", action="store_true"
    )
    parser.add_argument(
        "--path", "-p", help="Set the root path for contract imports", action="append", dest="paths"
    )
    parser.add_argument("-o", help="Set the output path", dest="output_path")
    parser.add_argument(
        "--experimental-codegen",
        help="The compiler use the new IR codegen. This is an experimental feature.",
        action="store_true",
        dest="experimental_codegen",
    )
    parser.add_argument("--enable-decimals", help="Enable decimals", action="store_true")
    parser.add_argument(
        "--disable-sys-path", help="Disable the use of sys.path", action="store_true"
    )

    args = parser.parse_args(argv)

    if args.traceback_limit is not None:
        sys.tracebacklimit = args.traceback_limit
    elif VYPER_TRACEBACK_LIMIT is not None:
        sys.tracebacklimit = VYPER_TRACEBACK_LIMIT
    elif args.verbose:
        sys.tracebacklimit = 1000
    else:
        # Python usually defaults sys.tracebacklimit to 1000.  We use a default
        # setting of zero so error printouts only include information about where
        # an error occurred in a Vyper source file.
        sys.tracebacklimit = 0

    if args.hex_ir:
        ir_node.AS_HEX_DEFAULT = True

    output_formats = tuple(uniq(args.format.split(",")))

    if args.base64 and output_formats != ("archive",):
        raise ValueError("Cannot use `--base64` except with `-f archive`")

    if args.base64:
        output_formats = ("archive_b64",)

    if args.no_optimize and args.optimize:
        raise ValueError("Cannot use `--no-optimize` and `--optimize` at the same time!")

    settings = Settings()

    if args.no_optimize:
        settings.optimize = OptimizationLevel.NONE
    elif args.optimize is not None:
        settings.optimize = OptimizationLevel.from_string(args.optimize)

    if args.evm_version:
        settings.evm_version = args.evm_version

    if args.experimental_codegen:
        settings.experimental_codegen = args.experimental_codegen

    if args.debug:
        settings.debug = args.debug

    if args.enable_decimals:
        settings.enable_decimals = args.enable_decimals

    if args.verbose:
        print(f"cli specified: `{settings}`", file=sys.stderr)

    include_sys_path = not args.disable_sys_path

    compiled = compile_files(
        args.input_files,
        output_formats,
        args.paths,
        include_sys_path,
        args.show_gas_estimates,
        settings,
        args.storage_layout,
        args.no_bytecode_metadata,
    )

    mode = "w"
    if output_formats == ("archive",):
        mode = "wb"

    if args.output_path:
        with open(args.output_path, mode) as f:
            _cli_helper(f, output_formats, compiled)
    else:
        # https://stackoverflow.com/a/54073813
        with os.fdopen(sys.stdout.fileno(), mode, closefd=False) as f:
            _cli_helper(f, output_formats, compiled)


def uniq(seq: Iterable[T]) -> Iterator[T]:
    """
    Yield unique items in ``seq`` in order.
    """
    seen: Set[T] = set()

    for x in seq:
        if x in seen:
            continue

        seen.add(x)
        yield x


def exc_handler(contract_path: ContractPath, exception: Exception) -> None:
    print(f"Error compiling: {contract_path}")
    raise exception


def get_search_paths(paths: list[str] = None, include_sys_path=True) -> list[Path]:
    # given `paths` input, get the full search path, including
    # the system search path.
    paths = paths or []

    # lowest precedence search path is always sys path
    # note python sys path uses opposite resolution order from us
    # (first in list is highest precedence; we give highest precedence
    # to the last in the list)
    search_paths = []
    if include_sys_path:
        search_paths = [Path(p) for p in reversed(sys.path)]

    if Path(".") not in search_paths:
        search_paths.append(Path("."))

    for p in paths:
        path = Path(p).resolve(strict=True)
        search_paths.append(path)

    return search_paths


def compile_files(
    input_files: list[str],
    output_formats: OutputFormats,
    paths: list[str] = None,
    include_sys_path: bool = True,
    show_gas_estimates: bool = False,
    settings: Optional[Settings] = None,
    storage_layout_paths: list[str] = None,
    no_bytecode_metadata: bool = False,
) -> dict:
    search_paths = get_search_paths(paths, include_sys_path)
    input_bundle = FilesystemInputBundle(search_paths)

    show_version = False
    if "combined_json" in output_formats:
        if len(output_formats) > 1:
            raise ValueError("If using combined_json it must be the only output format requested")
        output_formats = combined_json_outputs
        show_version = True

    # formats which can only be requested as a single output format
    for c in ("solc_json", "archive"):
        if c in output_formats and len(output_formats) > 1:
            raise ValueError(f"If using {c} it must be the only output format requested")

    translate_map = {
        "abi_python": "abi",
        "json": "abi",
        "ast": "ast_dict",
        "annotated_ast": "annotated_ast_dict",
        "ir_json": "ir_dict",
    }
    final_formats = [translate_map.get(i, i) for i in output_formats]

    if storage_layout_paths:
        if len(storage_layout_paths) != len(input_files):
            raise ValueError(
                f"provided {len(storage_layout_paths)} storage "
                f"layouts, but {len(input_files)} source files"
            )

    ret: dict[Any, Any] = {}
    if show_version:
        ret["version"] = vyper.__version__

    for file_name in input_files:
        file_path = Path(file_name)

        try:
            # try to compile in zipfile mode if it's a zip file, falling back
            # to regular mode if it's not.
            # we allow this instead of requiring a different mode (like
            # `--zip`) so that verifier pipelines do not need a different
            # workflow for archive files and single-file contracts.
            output = compile_from_zip(file_name, output_formats, settings, no_bytecode_metadata)
            ret[file_path] = output
            continue
        except NotZipInput:
            pass

        # note compile_from_zip also reads the file contents, so this
        # is slightly inefficient (and also maybe allows for some very
        # rare, strange race conditions if the file changes in between
        # the two reads).
        file = input_bundle.load_file(file_path)
        assert isinstance(file, FileInput)  # mypy hint

        storage_layout_override = None
        if storage_layout_paths:
            storage_file_path = storage_layout_paths.pop(0)
            with open(storage_file_path) as sfh:
                storage_layout_override = json.load(sfh)

        output = vyper.compile_from_file_input(
            file,
            input_bundle=input_bundle,
            output_formats=final_formats,
            exc_handler=exc_handler,
            settings=settings,
            storage_layout_override=storage_layout_override,
            show_gas_estimates=show_gas_estimates,
            no_bytecode_metadata=no_bytecode_metadata,
        )

        ret[file_path] = output

    return ret


if __name__ == "__main__":
    _parse_args(sys.argv[1:])



import pytest

from vyper.exceptions import FunctionDeclarationException


def test_constant_test(get_contract):
    constant_test = """
@external
@view
def foo() -> int128:
    return 5
    """

    c = get_contract(constant_test)
    assert c.foo() == 5

    print("Passed constant function test")


@pytest.mark.requires_evm_version("cancun")
def test_transient_test(get_contract):
    code = """
x: transient(uint256)

@external
@view
def foo() -> uint256:
    return self.x
    """
    c = get_contract(code)
    assert c.foo() == 0


def test_invalid_constant_and_payable(get_contract, assert_compile_failed):
    code = """
@external
@payable
@view
def foo() -> num:
    return 5
"""
    assert_compile_failed(lambda: get_contract(code), FunctionDeclarationException)



from vyper import compiler


def test_self_balance(env, get_contract):
    code = """
@external
@view
def get_balance() -> uint256:
    a: uint256 = self.balance
    return a

@external
@payable
def __default__():
    pass
    """
    opcodes = compiler.compile_code(code, output_formats=["opcodes"])["opcodes"]
    assert "SELFBALANCE" in opcodes

    c = get_contract(code)
    env.set_balance(env.deployer, 1337)
    env.message_call(c.address, value=1337)

    assert c.get_balance() == 1337
    assert env.get_balance(c.address) == 1337
    assert env.get_balance(env.deployer) == 0



from math import ceil

from vyper.codegen.core import bytes_data_ptr, ensure_in_memory, get_bytearray_length
from vyper.codegen.ir_node import IRnode
from vyper.exceptions import CompilerPanic
from vyper.semantics.types.bytestrings import _BytestringT
from vyper.semantics.types.shortcuts import BYTES32_T
from vyper.utils import SHA3_BASE, SHA3_PER_WORD, MemoryPositions


def _check_byteslike(typ):
    if not isinstance(typ, _BytestringT) and typ != BYTES32_T:  # pragma: nocover
        # NOTE this may be checked at a higher level, but just be safe
        raise CompilerPanic("keccak256 only accepts bytes-like objects")


def _gas_bound(num_words):
    return SHA3_BASE + num_words * SHA3_PER_WORD


def keccak256_helper(to_hash, context):
    _check_byteslike(to_hash.typ)

    # Can hash bytes32 objects
    # TODO: Want to generalize to all bytes_M
    if to_hash.typ == BYTES32_T:
        return IRnode.from_list(
            [
                "seq",
                ["mstore", MemoryPositions.FREE_VAR_SPACE, to_hash],
                ["sha3", MemoryPositions.FREE_VAR_SPACE, 32],
            ],
            typ=BYTES32_T,
            add_gas_estimate=_gas_bound(1),
        )

    to_hash = ensure_in_memory(to_hash, context)

    with to_hash.cache_when_complex("buf") as (b1, to_hash):
        data = bytes_data_ptr(to_hash)
        len_ = get_bytearray_length(to_hash)
        return b1.resolve(
            IRnode.from_list(
                ["sha3", data, len_],
                typ=BYTES32_T,
                annotation="keccak256",
                add_gas_estimate=_gas_bound(ceil(to_hash.typ.maxlen / 32)),
            )
        )



import pytest

from vyper import compiler
from vyper.exceptions import TypeMismatch, UndeclaredDefinition

type_fail_list = [
    """
@external
def foo():
    x: bytes32 = keccak256(3)
    """
]


@pytest.mark.parametrize("bad_code", type_fail_list)
def test_block_type_fail(bad_code):
    with pytest.raises(TypeMismatch):
        compiler.compile_code(bad_code)


structure_fail_list = [
    """
@external
def foo():
    x: bytes32 = sha3("moose")
    """,
    """
@external
def foo():
    x: bytes32 = sha3(0x1234567890123456789012345678901234567890123456789012345678901234)
    """,
]


@pytest.mark.parametrize("bad_code", structure_fail_list)
def test_block_structure_fail(bad_code):
    with pytest.raises(UndeclaredDefinition):
        compiler.compile_code(bad_code)


valid_list = [
    """
@external
def foo():
    x: bytes32 = keccak256("moose")
    """,
    """
@external
def foo():
    x: bytes32 = keccak256(0x1234567890123456789012345678901234567890123456789012345678901234)
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_block_success(good_code):
    assert compiler.compile_code(good_code) is not None



import pytest

from vyper.utils import annotate_source_code, indent

TEST_TEXT = """
test
lines
to
indent
"""[
    1:-1
]


def test_indent_indents_text():
    assert (
        indent(TEST_TEXT, indent_chars="-", level=1)
        == """
-test
-lines
-to
-indent
"""[
            1:-1
        ]
    )
    assert (
        indent(TEST_TEXT, indent_chars=" ", level=4)
        == """
    test
    lines
    to
    indent
"""[
            1:-1
        ]
    )
    assert (
        indent(TEST_TEXT, indent_chars=[" ", "*", "-", "="], level=4)
        == """
    test
****lines
----to
====indent
"""[
            1:-1
        ]
    )


def test_indent_raises_value_errors():
    with pytest.raises(ValueError, match="Must provide indentation chars for each line"):
        indent(TEST_TEXT, indent_chars=[" "], level=1)

    with pytest.raises(ValueError, match="Unrecognized indentation characters value"):
        indent(TEST_TEXT, indent_chars=None, level=1)  # type: ignore


TEST_SOURCE_CODE = r"""
# Attempts to display the line and column of violating code.
class ParserException(Exception):
    def __init__(self, message='Error Message not found.', item=None):
        self.message = message
        self.lineno = None
        self.col_offset = None

        if isinstance(item, tuple):  # is a position.
            self.lineno, self.col_offset = item
        elif item and hasattr(item, 'lineno'):
            self.set_err_pos(item.lineno, item.col_offset)
            if hasattr(item, 'source_code'):
                self.source_code = item.source_code.splitlines()

    def set_err_pos(self, lineno, col_offset):
        if not self.lineno:
            self.lineno = lineno

            if not self.col_offset:
                self.col_offset = col_offset

    def __str__(self):
        output = self.message

        if self.lineno and hasattr(self, 'source_code'):

            output = f'line {self.lineno}: {output}\n{self.source_code[self.lineno -1]}'

            if self.col_offset:
                col = '-' * self.col_offset + '^'
                output += '\n' + col

        elif self.lineno is not None and self.col_offset is not None:
            output = f'line {self.lineno}:{self.col_offset} {output}'

        return output
"""[
    1:-1
]


def test_annotate_source_code_marks_positions_in_source_code():
    annotation = annotate_source_code(
        TEST_SOURCE_CODE, 22, col_offset=16, context_lines=0, line_numbers=False
    )
    assert (
        annotation
        == r"""
    def __str__(self):
----------------^
"""[
            1:-1
        ]
    )

    annotation = annotate_source_code(
        TEST_SOURCE_CODE, 22, col_offset=15, context_lines=1, line_numbers=False
    )
    assert (
        annotation
        == r"""

    def __str__(self):
---------------^
        output = self.message
"""[
            1:-1
        ]
    )

    annotation = annotate_source_code(
        TEST_SOURCE_CODE, 22, col_offset=20, context_lines=2, line_numbers=False
    )
    assert (
        annotation
        == r"""
                self.col_offset = col_offset

    def __str__(self):
--------------------^
        output = self.message

"""[
            1:-1
        ]
    )

    annotation = annotate_source_code(
        TEST_SOURCE_CODE, 1, col_offset=5, context_lines=3, line_numbers=True
    )
    assert (
        annotation
        == r"""
---> 1 # Attempts to display the line and column of violating code.
------------^
     2 class ParserException(Exception):
     3     def __init__(self, message='Error Message not found.', item=None):
     4         self.message = message
"""[
            1:-1
        ]
    )

    annotation = annotate_source_code(
        TEST_SOURCE_CODE, 36, col_offset=8, context_lines=4, line_numbers=True
    )
    assert (
        annotation
        == r"""
     32
     33         elif self.lineno is not None and self.col_offset is not None:
     34             output = f'line {self.lineno}:{self.col_offset} {output}'
     35
---> 36         return output
----------------^
"""[
            1:-1
        ]
    )

    annotation = annotate_source_code(
        TEST_SOURCE_CODE, 15, col_offset=8, context_lines=11, line_numbers=True
    )
    assert (
        annotation
        == r"""
      4         self.message = message
      5         self.lineno = None
      6         self.col_offset = None
      7
      8         if isinstance(item, tuple):  # is a position.
      9             self.lineno, self.col_offset = item
     10         elif item and hasattr(item, 'lineno'):
     11             self.set_err_pos(item.lineno, item.col_offset)
     12             if hasattr(item, 'source_code'):
     13                 self.source_code = item.source_code.splitlines()
     14
---> 15     def set_err_pos(self, lineno, col_offset):
----------------^
     16         if not self.lineno:
     17             self.lineno = lineno
     18
     19             if not self.col_offset:
     20                 self.col_offset = col_offset
     21
     22     def __str__(self):
     23         output = self.message
     24
     25         if self.lineno and hasattr(self, 'source_code'):
     26
"""[
            1:-1
        ]
    )

    annotation = annotate_source_code(
        TEST_SOURCE_CODE, 15, col_offset=None, context_lines=3, line_numbers=True
    )
    assert (
        annotation
        == r"""
     12             if hasattr(item, 'source_code'):
     13                 self.source_code = item.source_code.splitlines()
     14
---> 15     def set_err_pos(self, lineno, col_offset):
     16         if not self.lineno:
     17             self.lineno = lineno
     18
"""[
            1:-1
        ]
    )

    annotation = annotate_source_code(
        TEST_SOURCE_CODE, 15, col_offset=None, context_lines=2, line_numbers=False
    )
    assert (
        annotation
        == r"""
                self.source_code = item.source_code.splitlines()

    def set_err_pos(self, lineno, col_offset):
        if not self.lineno:
            self.lineno = lineno
"""[
            1:-1
        ]
    )


@pytest.mark.parametrize("bad_lineno", (-100, -1, 0, 45, 1000))
def test_annotate_source_code_raises_value_errors(bad_lineno):
    with pytest.raises(ValueError, match="Line number is out of range"):
        annotate_source_code(TEST_SOURCE_CODE, bad_lineno)



from typing import Dict, List, Union

from vyper.ast import nodes as vy_ast
from vyper.exceptions import CompilerPanic


def ast_to_dict(ast_struct: Union[vy_ast.VyperNode, List]) -> Union[Dict, List]:
    """
    Converts a Vyper AST node, or list of nodes, into a dictionary suitable for
    output to the user.
    """
    if isinstance(ast_struct, vy_ast.VyperNode):
        return ast_struct.to_dict()
    elif isinstance(ast_struct, list):
        return [i.to_dict() for i in ast_struct]
    else:
        raise CompilerPanic(f'Unknown Vyper AST node provided: "{type(ast_struct)}".')


def dict_to_ast(ast_struct: Union[Dict, List]) -> Union[vy_ast.VyperNode, List]:
    """
    Converts an AST dict, or list of dicts, into Vyper AST node objects.
    """
    if isinstance(ast_struct, dict):
        return vy_ast.get_node(ast_struct)
    if isinstance(ast_struct, list):
        return [vy_ast.get_node(i) for i in ast_struct]
    raise CompilerPanic(f'Unknown ast_struct provided: "{type(ast_struct)}".')



import functools
import re
from typing import Optional

from vyper.codegen.ir_node import IRnode
from vyper.evm.opcodes import get_opcodes
from vyper.utils import MemoryPositions
from vyper.venom.basicblock import (
    IRBasicBlock,
    IRInstruction,
    IRLabel,
    IRLiteral,
    IROperand,
    IRVariable,
)
from vyper.venom.context import IRContext
from vyper.venom.function import IRFunction

# Instructions that are mapped to their inverse
INVERSE_MAPPED_IR_INSTRUCTIONS = {"ne": "eq", "le": "gt", "sle": "sgt", "ge": "lt", "sge": "slt"}

# Instructions that have a direct EVM opcode equivalent and can
# be passed through to the EVM assembly without special handling
PASS_THROUGH_INSTRUCTIONS = frozenset(
    [
        # binary instructions
        "eq",
        "gt",
        "lt",
        "slt",
        "sgt",
        "shr",
        "shl",
        "sar",
        "or",
        "xor",
        "and",
        "add",
        "sub",
        "mul",
        "div",
        "smul",
        "sdiv",
        "mod",
        "smod",
        "exp",
        "sha3",
        "sha3_64",
        "signextend",
        "chainid",
        "basefee",
        "blobhash",
        "blobbasefee",
        "timestamp",
        "blockhash",
        "caller",
        "selfbalance",
        "calldatasize",
        "callvalue",
        "address",
        "origin",
        "codesize",
        "gas",
        "gasprice",
        "gaslimit",
        "returndatasize",
        "mload",
        "iload",
        "istore",
        "sload",
        "sstore",
        "tload",
        "tstore",
        "coinbase",
        "number",
        "prevrandao",
        "difficulty",
        "iszero",
        "not",
        "calldataload",
        "extcodesize",
        "extcodehash",
        "balance",
        "msize",
        "invalid",
        "stop",
        "selfdestruct",
        "assert",
        "assert_unreachable",
        "exit",
        "calldatacopy",
        "mcopy",
        "extcodecopy",
        "codecopy",
        "returndatacopy",
        "revert",
        "create",
        "create2",
        "addmod",
        "mulmod",
        "call",
        "delegatecall",
        "staticcall",
    ]
)

NOOP_INSTRUCTIONS = frozenset(["pass", "cleanup_repeat", "var_list", "unique_symbol"])

SymbolTable = dict[str, Optional[IROperand]]
_global_symbols: SymbolTable = {}
MAIN_ENTRY_LABEL_NAME = "__main_entry"


# convert IRnode directly to venom
def ir_node_to_venom(ir: IRnode) -> IRContext:
    global _global_symbols
    _global_symbols = {}

    ctx = IRContext()
    fn = ctx.create_function(MAIN_ENTRY_LABEL_NAME)

    _convert_ir_bb(fn, ir, {})

    ctx.chain_basic_blocks()

    return ctx


def _append_jmp(fn: IRFunction, label: IRLabel) -> None:
    bb = fn.get_basic_block()
    if bb.is_terminated:
        bb = IRBasicBlock(fn.ctx.get_next_label("jmp_target"), fn)
        fn.append_basic_block(bb)

    bb.append_instruction("jmp", label)


def _new_block(fn: IRFunction) -> None:
    bb = IRBasicBlock(fn.ctx.get_next_label(), fn)
    fn.append_basic_block(bb)


def _append_return_args(fn: IRFunction, ofst: int = 0, size: int = 0):
    bb = fn.get_basic_block()
    if bb.is_terminated:
        bb = IRBasicBlock(fn.ctx.get_next_label("exit_to"), fn)
        fn.append_basic_block(bb)
    ret_ofst = IRVariable("ret_ofst")
    ret_size = IRVariable("ret_size")
    bb.append_instruction("store", ofst, ret=ret_ofst)
    bb.append_instruction("store", size, ret=ret_size)


def _handle_self_call(fn: IRFunction, ir: IRnode, symbols: SymbolTable) -> Optional[IRVariable]:
    setup_ir = ir.args[1]
    goto_ir = [ir for ir in ir.args if ir.value == "goto"][0]
    target_label = goto_ir.args[0].value  # goto
    return_buf_ir = goto_ir.args[1]  # return buffer
    ret_args: list[IROperand] = [IRLabel(target_label)]  # type: ignore

    if setup_ir != goto_ir:
        _convert_ir_bb(fn, setup_ir, symbols)

    return_buf = _convert_ir_bb(fn, return_buf_ir, symbols)

    bb = fn.get_basic_block()
    if len(goto_ir.args) > 2:
        ret_args.append(return_buf)  # type: ignore

    bb.append_invoke_instruction(ret_args, returns=False)  # type: ignore

    return return_buf


def _handle_internal_func(
    fn: IRFunction, ir: IRnode, does_return_data: bool, symbols: SymbolTable
) -> IRFunction:
    fn = fn.ctx.create_function(ir.args[0].args[0].value)
    bb = fn.get_basic_block()

    # return buffer
    if does_return_data:
        symbols["return_buffer"] = bb.append_instruction("param")
        bb.instructions[-1].annotation = "return_buffer"

    # return address
    symbols["return_pc"] = bb.append_instruction("param")
    bb.instructions[-1].annotation = "return_pc"

    _convert_ir_bb(fn, ir.args[0].args[2], symbols)

    return fn


def _convert_ir_simple_node(
    fn: IRFunction, ir: IRnode, symbols: SymbolTable
) -> Optional[IRVariable]:
    # execute in order
    args = _convert_ir_bb_list(fn, ir.args, symbols)
    # reverse output variables for stack
    args.reverse()
    return fn.get_basic_block().append_instruction(ir.value, *args)  # type: ignore


_break_target: Optional[IRBasicBlock] = None
_continue_target: Optional[IRBasicBlock] = None


def _convert_ir_bb_list(fn, ir, symbols):
    ret = []
    for ir_node in ir:
        venom = _convert_ir_bb(fn, ir_node, symbols)
        ret.append(venom)
    return ret


current_func = None
var_list: list[str] = []


def pop_source_on_return(func):
    @functools.wraps(func)
    def pop_source(*args, **kwargs):
        fn = args[0]
        ret = func(*args, **kwargs)
        fn.pop_source()
        return ret

    return pop_source


@pop_source_on_return
def _convert_ir_bb(fn, ir, symbols):
    assert isinstance(ir, IRnode), ir
    global _break_target, _continue_target, current_func, var_list, _global_symbols

    ctx = fn.ctx
    fn.push_source(ir)

    if ir.value in INVERSE_MAPPED_IR_INSTRUCTIONS:
        org_value = ir.value
        ir.value = INVERSE_MAPPED_IR_INSTRUCTIONS[ir.value]
        new_var = _convert_ir_simple_node(fn, ir, symbols)
        ir.value = org_value
        return fn.get_basic_block().append_instruction("iszero", new_var)
    elif ir.value in PASS_THROUGH_INSTRUCTIONS:
        return _convert_ir_simple_node(fn, ir, symbols)
    elif ir.value == "return":
        fn.get_basic_block().append_instruction(
            "return", IRVariable("ret_size"), IRVariable("ret_ofst")
        )
    elif ir.value == "deploy":
        ctx.ctor_mem_size = ir.args[0].value
        ctx.immutables_len = ir.args[2].value
        return None
    elif ir.value == "seq":
        if len(ir.args) == 0:
            return None
        if ir.is_self_call:
            return _handle_self_call(fn, ir, symbols)
        elif ir.args[0].value == "label":
            current_func = ir.args[0].args[0].value
            is_external = current_func.startswith("external")
            is_internal = current_func.startswith("internal")
            if is_internal or len(re.findall(r"external.*__init__\(.*_deploy", current_func)) > 0:
                # Internal definition
                var_list = ir.args[0].args[1]
                does_return_data = IRnode.from_list(["return_buffer"]) in var_list.args
                _global_symbols = {}
                symbols = {}
                new_fn = _handle_internal_func(fn, ir, does_return_data, symbols)
                for ir_node in ir.args[1:]:
                    ret = _convert_ir_bb(new_fn, ir_node, symbols)

                return ret
            elif is_external:
                _global_symbols = {}
                ret = _convert_ir_bb(fn, ir.args[0], symbols)
                _append_return_args(fn)
        else:
            bb = fn.get_basic_block()
            if bb.is_terminated:
                bb = IRBasicBlock(ctx.get_next_label("seq"), fn)
                fn.append_basic_block(bb)
            ret = _convert_ir_bb(fn, ir.args[0], symbols)

        for ir_node in ir.args[1:]:
            ret = _convert_ir_bb(fn, ir_node, symbols)

        return ret
    elif ir.value == "if":
        cond = ir.args[0]

        # convert the condition
        cont_ret = _convert_ir_bb(fn, cond, symbols)
        cond_block = fn.get_basic_block()

        saved_global_symbols = _global_symbols.copy()

        then_block = IRBasicBlock(ctx.get_next_label("then"), fn)
        else_block = IRBasicBlock(ctx.get_next_label("else"), fn)

        # convert "then"
        cond_symbols = symbols.copy()
        fn.append_basic_block(then_block)
        then_ret_val = _convert_ir_bb(fn, ir.args[1], cond_symbols)
        if isinstance(then_ret_val, IRLiteral):
            then_ret_val = fn.get_basic_block().append_instruction("store", then_ret_val)

        then_block_finish = fn.get_basic_block()

        # convert "else"
        cond_symbols = symbols.copy()
        _global_symbols = saved_global_symbols.copy()
        fn.append_basic_block(else_block)
        else_ret_val = None
        if len(ir.args) == 3:
            else_ret_val = _convert_ir_bb(fn, ir.args[2], cond_symbols)
            if isinstance(else_ret_val, IRLiteral):
                assert isinstance(else_ret_val.value, int)  # help mypy
                else_ret_val = fn.get_basic_block().append_instruction("store", else_ret_val)

        else_block_finish = fn.get_basic_block()

        # finish the condition block
        cond_block.append_instruction("jnz", cont_ret, then_block.label, else_block.label)

        # exit bb
        exit_bb = IRBasicBlock(ctx.get_next_label("if_exit"), fn)
        fn.append_basic_block(exit_bb)

        if_ret = fn.get_next_variable()
        if then_ret_val is not None and else_ret_val is not None:
            then_block_finish.append_instruction("store", then_ret_val, ret=if_ret)
            else_block_finish.append_instruction("store", else_ret_val, ret=if_ret)

        if not else_block_finish.is_terminated:
            else_block_finish.append_instruction("jmp", exit_bb.label)

        if not then_block_finish.is_terminated:
            then_block_finish.append_instruction("jmp", exit_bb.label)

        _global_symbols = saved_global_symbols

        return if_ret

    elif ir.value == "with":
        ret = _convert_ir_bb(fn, ir.args[1], symbols)  # initialization

        ret = fn.get_basic_block().append_instruction("store", ret)

        sym = ir.args[0]
        with_symbols = symbols.copy()
        with_symbols[sym.value] = ret

        return _convert_ir_bb(fn, ir.args[2], with_symbols)  # body

    elif ir.value == "goto":
        _append_jmp(fn, IRLabel(ir.args[0].value))
    elif ir.value == "djump":
        args = [_convert_ir_bb(fn, ir.args[0], symbols)]
        for target in ir.args[1:]:
            args.append(IRLabel(target.value))
        fn.get_basic_block().append_instruction("djmp", *args)
        _new_block(fn)
    elif ir.value == "set":
        sym = ir.args[0]
        arg_1 = _convert_ir_bb(fn, ir.args[1], symbols)
        fn.get_basic_block().append_instruction("store", arg_1, ret=symbols[sym.value])
    elif ir.value == "symbol":
        return IRLabel(ir.args[0].value, True)
    elif ir.value == "data":
        label = IRLabel(ir.args[0].value)
        ctx.append_data("dbname", [label])
        for c in ir.args[1:]:
            if isinstance(c, int):
                assert 0 <= c <= 255, "data with invalid size"
                ctx.append_data("db", [c])  # type: ignore
            elif isinstance(c.value, bytes):
                ctx.append_data("db", [c.value])  # type: ignore
            elif isinstance(c, IRnode):
                data = _convert_ir_bb(fn, c, symbols)
                ctx.append_data("db", [data])  # type: ignore
    elif ir.value == "label":
        label = IRLabel(ir.args[0].value, True)
        bb = fn.get_basic_block()
        if not bb.is_terminated:
            bb.append_instruction("jmp", label)
        bb = IRBasicBlock(label, fn)
        fn.append_basic_block(bb)
        code = ir.args[2]
        if code.value == "pass":
            bb.append_instruction("exit")
        else:
            _convert_ir_bb(fn, code, symbols)
    elif ir.value == "exit_to":
        args = _convert_ir_bb_list(fn, ir.args[1:], symbols)
        var_list = args
        _append_return_args(fn, *var_list)
        bb = fn.get_basic_block()
        if bb.is_terminated:
            bb = IRBasicBlock(ctx.get_next_label("exit_to"), fn)
            fn.append_basic_block(bb)
        bb = fn.get_basic_block()

        label = IRLabel(ir.args[0].value)
        if label.value == "return_pc":
            label = symbols.get("return_pc")
            bb.append_instruction("ret", label)
        else:
            bb.append_instruction("jmp", label)

    elif ir.value == "dload":
        arg_0 = _convert_ir_bb(fn, ir.args[0], symbols)
        bb = fn.get_basic_block()
        src = bb.append_instruction("add", arg_0, IRLabel("code_end"))

        bb.append_instruction("dloadbytes", 32, src, MemoryPositions.FREE_VAR_SPACE)
        return bb.append_instruction("mload", MemoryPositions.FREE_VAR_SPACE)

    elif ir.value == "dloadbytes":
        dst, src_offset, len_ = _convert_ir_bb_list(fn, ir.args, symbols)

        bb = fn.get_basic_block()
        src = bb.append_instruction("add", src_offset, IRLabel("code_end"))
        bb.append_instruction("dloadbytes", len_, src, dst)
        return None

    elif ir.value == "mstore":
        # some upstream code depends on reversed order of evaluation --
        # to fix upstream.
        val, ptr = _convert_ir_bb_list(fn, reversed(ir.args), symbols)

        return fn.get_basic_block().append_instruction("mstore", val, ptr)

    elif ir.value == "ceil32":
        x = ir.args[0]
        expanded = IRnode.from_list(["and", ["add", x, 31], ["not", 31]])
        return _convert_ir_bb(fn, expanded, symbols)
    elif ir.value == "select":
        cond, a, b = ir.args
        expanded = IRnode.from_list(
            [
                "with",
                "cond",
                cond,
                [
                    "with",
                    "a",
                    a,
                    ["with", "b", b, ["xor", "b", ["mul", "cond", ["xor", "a", "b"]]]],
                ],
            ]
        )
        return _convert_ir_bb(fn, expanded, symbols)
    elif ir.value == "repeat":

        def emit_body_blocks():
            global _break_target, _continue_target, _global_symbols
            old_targets = _break_target, _continue_target
            _break_target, _continue_target = exit_block, incr_block
            saved_global_symbols = _global_symbols.copy()
            _convert_ir_bb(fn, body, symbols.copy())
            _break_target, _continue_target = old_targets
            _global_symbols = saved_global_symbols

        sym = ir.args[0]
        start, end, _ = _convert_ir_bb_list(fn, ir.args[1:4], symbols)

        assert ir.args[3].is_literal, "repeat bound expected to be literal"

        bound = ir.args[3].value
        if (
            isinstance(end, IRLiteral)
            and isinstance(start, IRLiteral)
            and end.value + start.value <= bound
        ):
            bound = None

        body = ir.args[4]

        entry_block = IRBasicBlock(ctx.get_next_label("repeat"), fn)
        cond_block = IRBasicBlock(ctx.get_next_label("condition"), fn)
        body_block = IRBasicBlock(ctx.get_next_label("body"), fn)
        incr_block = IRBasicBlock(ctx.get_next_label("incr"), fn)
        exit_block = IRBasicBlock(ctx.get_next_label("exit"), fn)

        bb = fn.get_basic_block()
        bb.append_instruction("jmp", entry_block.label)
        fn.append_basic_block(entry_block)

        counter_var = entry_block.append_instruction("store", start)
        symbols[sym.value] = counter_var
        end = entry_block.append_instruction("add", start, end)
        if bound:
            bound = entry_block.append_instruction("add", start, bound)
        entry_block.append_instruction("jmp", cond_block.label)

        xor_ret = cond_block.append_instruction("xor", counter_var, end)
        cont_ret = cond_block.append_instruction("iszero", xor_ret)
        fn.append_basic_block(cond_block)

        fn.append_basic_block(body_block)
        if bound:
            xor_ret = body_block.append_instruction("xor", counter_var, bound)
            body_block.append_instruction("assert", xor_ret)

        emit_body_blocks()
        body_end = fn.get_basic_block()
        if body_end.is_terminated is False:
            body_end.append_instruction("jmp", incr_block.label)

        fn.append_basic_block(incr_block)
        incr_block.insert_instruction(
            IRInstruction("add", [counter_var, IRLiteral(1)], counter_var)
        )
        incr_block.append_instruction("jmp", cond_block.label)

        fn.append_basic_block(exit_block)

        cond_block.append_instruction("jnz", cont_ret, exit_block.label, body_block.label)
    elif ir.value == "break":
        assert _break_target is not None, "Break with no break target"
        fn.get_basic_block().append_instruction("jmp", _break_target.label)
        fn.append_basic_block(IRBasicBlock(ctx.get_next_label(), fn))
    elif ir.value == "continue":
        assert _continue_target is not None, "Continue with no contrinue target"
        fn.get_basic_block().append_instruction("jmp", _continue_target.label)
        fn.append_basic_block(IRBasicBlock(ctx.get_next_label(), fn))
    elif ir.value in NOOP_INSTRUCTIONS:
        pass
    elif isinstance(ir.value, str) and ir.value.startswith("log"):
        args = reversed(_convert_ir_bb_list(fn, ir.args, symbols))
        topic_count = int(ir.value[3:])
        assert topic_count >= 0 and topic_count <= 4, "invalid topic count"
        fn.get_basic_block().append_instruction("log", topic_count, *args)
    elif isinstance(ir.value, str) and ir.value.upper() in get_opcodes():
        _convert_ir_opcode(fn, ir, symbols)
    elif isinstance(ir.value, str):
        if ir.value.startswith("$alloca") and ir.value not in _global_symbols:
            alloca = ir.passthrough_metadata["alloca"]
            ptr = fn.get_basic_block().append_instruction("alloca", alloca.offset, alloca.size)
            _global_symbols[ir.value] = ptr
        elif ir.value.startswith("$palloca") and ir.value not in _global_symbols:
            alloca = ir.passthrough_metadata["alloca"]
            ptr = fn.get_basic_block().append_instruction("store", alloca.offset)
            _global_symbols[ir.value] = ptr

        return _global_symbols.get(ir.value) or symbols.get(ir.value)
    elif ir.is_literal:
        return IRLiteral(ir.value)
    else:
        raise Exception(f"Unknown IR node: {ir}")

    return None


def _convert_ir_opcode(fn: IRFunction, ir: IRnode, symbols: SymbolTable) -> None:
    opcode = ir.value.upper()  # type: ignore
    inst_args = []
    for arg in ir.args:
        if isinstance(arg, IRnode):
            inst_args.append(_convert_ir_bb(fn, arg, symbols))
    fn.get_basic_block().append_instruction(opcode, *inst_args)



from vyper.exceptions import InvalidAttribute


def test_multi_setter_test(get_contract):
    multi_setter_test = """
dog: int128[3]
bar: int128[3][3]
@external
def foo() -> int128:
    self.dog = [1, 2, 3]
    return(self.dog[0] + self.dog[1] * 10 + self.dog[2] * 100)

@external
def fop() -> int128:
    self.bar[0] = [1, 2, 3]
    self.bar[1] = [4, 5, 6]
    return self.bar[0][0] + self.bar[0][1] * 10 + self.bar[0][2] * 100 + \
        self.bar[1][0] * 1000 + self.bar[1][1] * 10000 + self.bar[1][2] * 100000

@external
def goo() -> int128:
    god: int128[3] = [1, 2, 3]
    return(god[0] + god[1] * 10 + god[2] * 100)

@external
def gop() -> int128: # Following a standard naming scheme; nothing to do with the US republican party  # noqa: E501
    gar: int128[3][3] = [[1, 2, 3], [4, 5, 6], [0, 0, 0]]
    return gar[0][0] + gar[0][1] * 10 + gar[0][2] * 100 + \
        gar[1][0] * 1000 + gar[1][1] * 10000 + gar[1][2] * 100000

@external
def hoo() -> int128:
    self.dog = empty(int128[3])
    return(self.dog[0] + self.dog[1] * 10 + self.dog[2] * 100)

@external
def hop() -> int128:
    self.bar[1] = empty(int128[3])
    return self.bar[0][0] + self.bar[0][1] * 10 + self.bar[0][2] * 100 + \
        self.bar[1][0] * 1000 + self.bar[1][1] * 10000 + self.bar[1][2] * 100000

@external
def joo() -> int128:
    god: int128[3] = [1, 2, 3]
    god = empty(int128[3])
    return(god[0] + god[1] * 10 + god[2] * 100)

@external
def jop() -> int128:
    gar: int128[3][3] = [[1, 2, 3], [4, 5, 6], [0, 0, 0]]
    gar[1] = empty(int128[3])
    return gar[0][0] + gar[0][1] * 10 + gar[0][2] * 100 + \
        gar[1][0] * 1000 + gar[1][1] * 10000 + gar[1][2] * 100000

    """

    c = get_contract(multi_setter_test)
    assert c.foo() == 321
    c.foo()
    assert c.fop() == 654321
    c.fop()
    assert c.goo() == 321
    assert c.gop() == 654321
    assert c.hoo() == 0
    assert c.hop() == 321
    assert c.joo() == 0
    assert c.jop() == 321
    print("Passed multi-setter literal test")


def test_multi_setter_struct_test(get_contract):
    multi_setter_struct_test = """
struct Dog:
    foo: int128
    bar: int128
struct Bar:
    a: int128
    b: int128
struct Z:
    foo: int128[3]
    bar: Bar[2]
struct Goo:
    foo: int128
    bar: int128
struct Zed:
    foo: int128[3]
    bar: Bar[2]
dog: Dog[3]
z: Z[2]

@external
def foo() -> int128:
    foo0: int128 = 1
    self.dog[0] = Dog(foo=foo0, bar=2)
    self.dog[1] = Dog(foo=3, bar=4)
    self.dog[2] = Dog(foo=5, bar=6)
    return self.dog[0].foo + self.dog[0].bar * 10 + self.dog[1].foo * 100 + \
        self.dog[1].bar * 1000 + self.dog[2].foo * 10000 + self.dog[2].bar * 100000

@external
def fop() -> int128:
    self.z = [Z(foo=[1, 2, 3], bar=[Bar(a=4, b=5), Bar(a=2, b=3)]),
              Z(foo=[6, 7, 8], bar=[Bar(a=9, b=1), Bar(a=7, b=8)])]
    return self.z[0].foo[0] + self.z[0].foo[1] * 10 + self.z[0].foo[2] * 100 + \
        self.z[0].bar[0].a * 1000 + \
        self.z[0].bar[0].b * 10000 + \
        self.z[0].bar[1].a * 100000 + \
        self.z[0].bar[1].b * 1000000 + \
        self.z[1].foo[0] * 10000000 + \
        self.z[1].foo[1] * 100000000 + \
        self.z[1].foo[2] * 1000000000 + \
        self.z[1].bar[0].a * 10000000000 + \
        self.z[1].bar[0].b * 100000000000 + \
        self.z[1].bar[1].a * 1000000000000 + \
        self.z[1].bar[1].b * 10000000000000

@external
def goo() -> int128:
    god: Goo[3] = [Goo(foo=1, bar=2), Goo(foo=3, bar=4), Goo(foo=5, bar=6)]
    return god[0].foo + god[0].bar * 10 + god[1].foo * 100 + \
        god[1].bar * 1000 + god[2].foo * 10000 + god[2].bar * 100000

@external
def gop() -> int128:
    zed: Zed[2] = [
        Zed(foo=[1, 2, 3], bar=[Bar(a=4, b=5), Bar(a=2, b=3)]),
        Zed(foo=[6, 7, 8], bar=[Bar(a=9, b=1), Bar(a=7, b=8)])
    ]
    return zed[0].foo[0] + zed[0].foo[1] * 10 + \
        zed[0].foo[2] * 100 + \
        zed[0].bar[0].a * 1000 + \
        zed[0].bar[0].b * 10000 + \
        zed[0].bar[1].a * 100000 + \
        zed[0].bar[1].b * 1000000 + \
        zed[1].foo[0] * 10000000 + \
        zed[1].foo[1] * 100000000 + \
        zed[1].foo[2] * 1000000000 + \
        zed[1].bar[0].a * 10000000000 + \
        zed[1].bar[0].b * 100000000000 + \
        zed[1].bar[1].a * 1000000000000 + \
        zed[1].bar[1].b * 10000000000000
    """

    c = get_contract(multi_setter_struct_test)
    assert c.foo() == 654321
    assert c.fop() == 87198763254321
    assert c.goo() == 654321
    assert c.gop() == 87198763254321


def test_struct_assignment_order(get_contract, assert_compile_failed):
    code = """
struct Foo:
    a: uint256
    b: uint256

@external
@view
def test2() -> uint256:
    foo: Foo = Foo(b=2, a=297)
    return foo.a
    """
    assert_compile_failed(lambda: get_contract(code), InvalidAttribute)


def test_type_converter_setter_test(get_contract):
    type_converter_setter_test = """
pap: decimal[2][2]

@external
def goo() -> int256:
    self.pap = [[1.0, 2.0], [3.0, 4.0]]
    return floor(
        self.pap[0][0] +
        self.pap[0][1] * 10.0 +
        self.pap[1][0] * 100.0 +
        self.pap[1][1] * 1000.0)
    """

    c = get_contract(type_converter_setter_test)
    assert c.goo() == 4321


def test_composite_setter_test(get_contract):
    composite_setter_test = """
struct C:
    c: int128
struct Mom:
    a: C[3]
    b: int128
mom: Mom
qoq: C

@external
def foo() -> int128:
    self.mom = Mom(a=[C(c=1), C(c=2), C(c=3)], b=4)
    non: C = C(c=5)
    self.mom.a[0] = non
    non = C(c=6)
    self.mom.a[2] = non
    return self.mom.a[0].c + self.mom.a[1].c * 10 + self.mom.a[2].c * 100 + self.mom.b * 1000

@external
def fop() -> int128:
    popp: Mom = Mom(a=[C(c=1), C(c=2), C(c=3)], b=4)
    self.qoq = C(c=5)
    popp.a[0] = self.qoq
    self.qoq = C(c=6)
    popp.a[2] = self.qoq
    return popp.a[0].c + popp.a[1].c * 10 + popp.a[2].c * 100 + popp.b * 1000

@external
def foq() -> int128:
    popp: Mom = Mom(a=[C(c=1), C(c=2), C(c=3)], b=4)
    popp.a[0] = empty(C)
    popp.a[2] = empty(C)
    return popp.a[0].c + popp.a[1].c * 10 + popp.a[2].c * 100 + popp.b * 1000
    """

    c = get_contract(composite_setter_test)
    assert c.foo() == 4625
    assert c.fop() == 4625
    assert c.foq() == 4020



import pytest

from vyper import compiler
from vyper.exceptions import InvalidLiteral, TypeMismatch

fail_list = [
    (
        """
bar: int128[3][3]
@external
def foo():
    self.bar = [[1, 2], [3, 4, 5], [6, 7, 8]]
    """,
        TypeMismatch,  # casting darray to sarray
    ),
    (
        """
bar: int128[3][3]
@external
def foo():
    self.bar = [[1, 2, 3], [4, 5, 6], [7.0, 8.0, 9.0]]
    """,
        InvalidLiteral,
    ),
    (
        """
@external
def foo() -> int128[2]:
    return [[1,2],[3,4]]
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo() -> int128[2][2]:
    return [1,2]
    """,
        TypeMismatch,
    ),
    (
        """
y: address[2][2]

@external
def foo(x: int128[2][2]) -> int128:
    self.y = x
    return 768
    """,
        TypeMismatch,
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_nested_list_fail(bad_code, exc):
    with pytest.raises(exc):
        compiler.compile_code(bad_code)


valid_list = [
    """
bar: int128[3][3]
@external
def foo():
    self.bar = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    """,
    """
bar: decimal[3][3]
@external
def foo():
    self.bar = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_nested_list_success(good_code):
    assert compiler.compile_code(good_code) is not None



import pytest

from vyper.compiler import compile_code
from vyper.exceptions import FunctionDeclarationException, StateAccessViolation


def test_pure_operation(get_contract):
    code = """
@pure
@external
def foo() -> int128:
    return 5
    """
    c = get_contract(code)
    assert c.foo() == 5


def test_pure_call(get_contract):
    code = """
@pure
@internal
def _foo() -> int128:
    return 5

@pure
@external
def foo() -> int128:
    return self._foo()
    """
    c = get_contract(code)
    assert c.foo() == 5


def test_pure_interface(get_contract):
    code1 = """
@pure
@external
def foo() -> int128:
    return 5
    """
    code2 = """
interface Foo:
    def foo() -> int128: pure

@pure
@external
def foo(a: address) -> int128:
    return staticcall Foo(a).foo()
    """
    c1 = get_contract(code1)
    c2 = get_contract(code2)
    assert c2.foo(c1.address) == 5


def test_invalid_envar_access(get_contract):
    code = """
@pure
@external
def foo() -> uint256:
    return chain.id
    """
    with pytest.raises(StateAccessViolation):
        compile_code(code)


def test_invalid_codesize_access(get_contract):
    code = """
@pure
@external
def foo(s: address) -> uint256:
    return s.codesize
    """
    with pytest.raises(StateAccessViolation):
        compile_code(code)


def test_invalid_state_access(get_contract, assert_compile_failed):
    code = """
x: uint256

@pure
@external
def foo() -> uint256:
    return self.x
    """
    with pytest.raises(StateAccessViolation):
        compile_code(code)


def test_invalid_immutable_access():
    code = """
COUNTER: immutable(uint256)

@deploy
def __init__():
    COUNTER = 1234

@pure
@external
def foo() -> uint256:
    return COUNTER
    """
    with pytest.raises(StateAccessViolation):
        compile_code(code)


def test_invalid_self_access():
    code = """
@pure
@external
def foo() -> address:
    return self
    """
    with pytest.raises(StateAccessViolation):
        compile_code(code)


def test_invalid_module_variable_access(make_input_bundle):
    lib1 = """
counter: uint256
    """
    code = """
import lib1
initializes: lib1

@pure
@external
def foo() -> uint256:
    return lib1.counter
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    with pytest.raises(StateAccessViolation):
        compile_code(code, input_bundle=input_bundle)


def test_invalid_module_immutable_access(make_input_bundle):
    lib1 = """
COUNTER: immutable(uint256)

@deploy
def __init__():
    COUNTER = 123
    """
    code = """
import lib1
initializes: lib1

@deploy
def __init__():
    lib1.__init__()

@pure
@external
def foo() -> uint256:
    return lib1.COUNTER
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    with pytest.raises(StateAccessViolation):
        compile_code(code, input_bundle=input_bundle)


def test_invalid_call():
    code = """
@view
@internal
def _foo() -> uint256:
    return 5

@pure
@external
def foo() -> uint256:
    return self._foo()  # Fails because of calling non-pure fn
    """
    with pytest.raises(StateAccessViolation):
        compile_code(code)


def test_type_in_pure(get_contract):
    code = """
@pure
@external
def _convert(x: bytes32) -> uint256:
    return convert(x, uint256)
    """
    c = get_contract(code)
    x = 123456
    bs = x.to_bytes(32, "big")
    assert x == c._convert(bs)


def test_invalid_conflicting_decorators():
    code = """
@pure
@external
@payable
def foo() -> uint256:
    return 5
    """
    with pytest.raises(FunctionDeclarationException):
        compile_code(code)


@pytest.mark.requires_evm_version("cancun")
def test_invalid_transient_access():
    code = """
x: transient(uint256)

@external
@pure
def foo() -> uint256:
    return self.x
    """
    with pytest.raises(StateAccessViolation):
        compile_code(code)



import pytest

from vyper.compiler import compile_code
from vyper.exceptions import CallViolation, StructureException

good = [
    # payable and nonpayable, "plain" (no assignment)
    """
interface Foo:
    def foo(): payable
    def bar() -> uint256: nonpayable

@external
def foo(f: Foo):
    extcall f.foo()
    extcall f.bar()
    """,
    """
interface Foo:
    def foo(): nonpayable
    def bar() -> uint256: nonpayable

@external
def foo(f: Foo):
    extcall f.foo()
    extcall f.bar()
    """,
    # payable and nonpayable, with assignment
    """
interface Foo:
    def foo() -> uint256: payable

@external
def foo(f: Foo):
    s: uint256 = extcall f.foo()
    """,
    """
interface Foo:
    def foo() -> uint256: nonpayable

@external
def foo(f: Foo):
    s: uint256 = extcall f.foo()
    """,
    # view and pure functions
    """
interface Foo:
    def foo() -> uint256: view

@external
def foo(f: Foo):
    s: uint256 = staticcall f.foo()
    """,
    # view and pure functions
    """
interface Foo:
    def foo() -> uint256: pure

@external
def foo(f: Foo):
    s: uint256 = staticcall f.foo()
    """,
    # TODO: tokenizer currently has issue with log+staticcall/extcall, e.g.
    # `log Bar(staticcall f.foo() + extcall f.bar())`
]


@pytest.mark.parametrize("code", good)
def test_good(code):
    _ = compile_code(code)


bad = [
    (
        """
interface Foo:
    def foo() -> uint256: pure

@internal
def bar(f: Foo):
    f.foo()
    """,
        CallViolation,
        "Calls to external pure functions must use the `staticcall` keyword.",
        "try `staticcall f.foo()`",
    ),
    (
        """
interface Foo:
    def foo() -> uint256: view

@internal
def bar(f: Foo):
    f.foo()
    """,
        CallViolation,
        "Calls to external view functions must use the `staticcall` keyword.",
        "try `staticcall f.foo()`",
    ),
    (
        """
interface Foo:
    def foo() -> uint256: nonpayable

@internal
def bar(f: Foo):
    f.foo()
    """,
        CallViolation,
        "Calls to external nonpayable functions must use the `extcall` keyword.",
        "try `extcall f.foo()`",
    ),
    (
        """
interface Foo:
    def foo() -> uint256: payable

@internal
def bar(f: Foo):
    f.foo()
    """,
        CallViolation,
        "Calls to external payable functions must use the `extcall` keyword.",
        "try `extcall f.foo()`",
    ),
    (
        """
interface Foo:
    def foo() -> uint256: nonpayable

@internal
def bar(f: Foo):
    s: uint256 = staticcall f.foo()
    """,
        CallViolation,
        "Calls to external nonpayable functions must use the `extcall` keyword.",
        "try `extcall f.foo()`",
    ),
    (
        """
interface Foo:
    def foo() -> uint256: view

@internal
def bar(f: Foo):
    s: uint256 = extcall f.foo()
    """,
        CallViolation,
        "Calls to external view functions must use the `staticcall` keyword.",
        "try `staticcall f.foo()`",
    ),
    (  # staticcall without assigning result disallowed
        """
interface Foo:
    def foo() -> uint256: view

@internal
def bar(f: Foo):
    staticcall f.foo()
    """,
        StructureException,
        "Expressions without assignment are disallowed",
        "did you mean to assign the result to a variable?",
    ),
    (
        """
interface Foo:
    def foo() -> uint256: view

@internal
def bar(f: Foo):
    staticcall f.foo()
    """,
        StructureException,
        "Expressions without assignment are disallowed",
        "did you mean to assign the result to a variable?",
    ),
    (
        """
@internal
def foo() -> uint256:
    return 1

@internal
def bar():
    s: uint256 = staticcall self.foo()
    """,
        CallViolation,
        "Calls to internal functions cannot use the `staticcall` keyword.",
        "remove the `staticcall` keyword",
    ),
    (
        """
@internal
def foo() -> uint256:
    return 1

@internal
def bar():
    s: uint256 = extcall self.foo()
    """,
        CallViolation,
        "Calls to internal functions cannot use the `extcall` keyword.",
        "remove the `extcall` keyword",
    ),
    (
        """
@internal
def foo():
    pass

@internal
def bar():
    extcall self.foo()
    """,
        CallViolation,
        "Calls to internal functions cannot use the `extcall` keyword.",
        "remove the `extcall` keyword",
    ),
    (
        """
@internal
def bar():
    extcall x
    """,
        StructureException,
        "`extcall` must be followed by a function call",
        "did you forget parentheses?",
    ),
    (
        """
@internal
def bar():
    staticcall x
    """,
        StructureException,
        "`staticcall` must be followed by a function call",
        "did you forget parentheses?",
    ),
    (  # test cannot call builtin
        """
@internal
def bar():
    extcall raw_call(msg.sender, b"")
    """,
        CallViolation,
        "cannot use `extcall` here!",
        "remove the `extcall` keyword",
    ),
    (  # test cannot call MemberFunctionT
        """
@internal
def bar():
    s: DynArray[uint256, 6] = []
    extcall s.pop()
    """,
        CallViolation,
        "cannot use `extcall` here!",
        "remove the `extcall` keyword",
    ),
    (  # test cannot extcall struct ctor
        """
struct Foo:
    x: uint256
@internal
def bar():
    s: Foo = extcall Foo(x=1)
    """,
        CallViolation,
        "cannot use `extcall` here!",
        "remove the `extcall` keyword",
    ),
    # maybe this test belongs in the logging tests
    (  # test cannot extcall log ctor
        """
event Foo:
    x: uint256
@internal
def bar():
    log extcall Foo(1)
    """,
        StructureException,
        "Log must call an event",
        None,
    ),
    (  # test cannot extcall event ctor
        """
event Foo:
    x: uint256
@internal
def bar():
    extcall Foo(1)
    """,
        StructureException,
        "To call an event you must use the `log` statement",
        None,
    ),
    (  # test cannot extcall interface ctor
        """
interface Foo:
    def foo(): nonpayable

@internal
def bar():
    extcall Foo(msg.sender)
    """,
        StructureException,
        "Function `type(interface Foo)` cannot be called without assigning the result",
        None,
    ),
]


@pytest.mark.parametrize("code,exc_type,msg,hint", bad)
def test_bad(code, exc_type, msg, hint):
    with pytest.raises(exc_type) as e:
        _ = compile_code(code)
    assert e.value._message == msg
    assert e.value._hint == hint



from collections import defaultdict

from vyper.exceptions import ExceptionList, InitializerException
from vyper.semantics.analysis.base import InitializesInfo, UsesInfo
from vyper.semantics.types.module import ModuleT


def validate_compilation_target(module_t: ModuleT):
    _validate_global_initializes_constraint(module_t)


def _collect_used_modules_r(module_t):
    ret: defaultdict[ModuleT, list[UsesInfo]] = defaultdict(list)

    for uses_decl in module_t.uses_decls:
        for used_module in uses_decl._metadata["uses_info"].used_modules:
            ret[used_module.module_t].append(uses_decl)

            # recurse
            used_modules = _collect_used_modules_r(used_module.module_t)
            for k, v in used_modules.items():
                ret[k].extend(v)

    # also recurse into modules used by initialized modules
    for i in module_t.initialized_modules:
        used_modules = _collect_used_modules_r(i.module_info.module_t)
        for k, v in used_modules.items():
            ret[k].extend(v)

    return ret


def _collect_initialized_modules_r(module_t, seen=None):
    seen: dict[ModuleT, InitializesInfo] = seen or {}

    # list of InitializedInfo
    initialized_infos = module_t.initialized_modules

    for i in initialized_infos:
        initialized_module_t = i.module_info.module_t
        if initialized_module_t in seen:
            seen_nodes = (i.node, seen[initialized_module_t].node)
            raise InitializerException(f"`{i.module_info.alias}` initialized twice!", *seen_nodes)
        seen[initialized_module_t] = i

        _collect_initialized_modules_r(initialized_module_t, seen)

    return seen


# validate that each module which is `used` in the import graph is
# `initialized`.
def _validate_global_initializes_constraint(module_t: ModuleT):
    all_used_modules = _collect_used_modules_r(module_t)
    all_initialized_modules = _collect_initialized_modules_r(module_t)

    err_list = ExceptionList()

    for u, uses in all_used_modules.items():
        if u not in all_initialized_modules:
            found_module = module_t.find_module_info(u)
            if found_module is not None:
                # TODO: do something about these constants
                if str(module_t) in ("<unknown>", "VyperContract.vy"):
                    module_str = "the top level of your main contract"
                else:
                    module_str = f"`{module_t}`"
                hint = f"add `initializes: {found_module.alias}` to {module_str}"
            else:
                # CMC 2024-02-06 is this actually reachable?
                hint = f"ensure `{module_t}` is imported in your main contract!"
            err_list.append(
                InitializerException(
                    f"module `{u}` is used but never initialized!", *uses, hint=hint
                )
            )

    err_list.raise_if_not_empty()



import pytest

from vyper import compile_code
from vyper.exceptions import ArgumentException, InvalidType, SyntaxException, TypeMismatch

fail_list = [
    (
        """
@external
def foo():
    x: Bytes[9] = raw_call(
        0x1234567890123456789012345678901234567890, b"cow", max_outsize=4, max_outsize=9
    )
    """,
        (SyntaxException, ArgumentException),
    ),
    (
        """
@external
@view
def foo(_addr: bytes4):
    # bytes4 instead of address
    raw_call(_addr, method_id("foo()"))
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo():
    # fails because raw_call without max_outsize does not return a value
    x: Bytes[9] = raw_call(0x1234567890123456789012345678901234567890, b"cow")
    """,
        InvalidType,
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_raw_call_fail(bad_code, exc):
    with pytest.raises(exc):
        compile_code(bad_code)


valid_list = [
    """
@external
def foo():
    x: Bytes[9] = raw_call(
        0x1234567890123456789012345678901234567890,
        b"cow",
        max_outsize=4,
        gas=595757
    )
    """,
    """
@external
def foo():
    x: Bytes[9] = raw_call(
        0x1234567890123456789012345678901234567890,
        b"cow",
        max_outsize=4,
        gas=595757,
        value=as_wei_value(9, "wei")
    )
    """,
    """
@external
def foo():
    x: Bytes[9] = raw_call(
        0x1234567890123456789012345678901234567890,
        b"cow",
        max_outsize=4,
        gas=595757,
        value=9
    )
    """,
    """
@external
def foo():
    raw_call(0x1234567890123456789012345678901234567890, b"cow")
    """,
    """
balances: HashMap[uint256,uint256]
@external
def foo():
    raw_call(
        0x1234567890123456789012345678901234567890,
        b"cow",
        value=self.balance - self.balances[0]
    )
    """,
    # test constants
    """
OUTSIZE: constant(uint256) = 4
REVERT_ON_FAILURE: constant(bool) = True
@external
def foo():
    x: Bytes[9] = raw_call(
        0x1234567890123456789012345678901234567890,
        b"cow",
        max_outsize=OUTSIZE,
        gas=595757,
        revert_on_failure=REVERT_ON_FAILURE
    )
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_raw_call_success(good_code):
    assert compile_code(good_code) is not None



import pytest

from vyper import compiler
from vyper.exceptions import TypeMismatch

fail_list = [
    (
        """
@external
def foo(inp: Bytes[10]) -> Bytes[2]:
    return slice(inp, 2, 3)
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo(inp: int128) -> Bytes[3]:
    return slice(inp, 2, 3)
    """,
        TypeMismatch,
    ),
    (
        """
@external
def foo(inp: Bytes[10]) -> Bytes[3]:
    return slice(inp, 4.0, 3)
    """,
        TypeMismatch,
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_slice_fail(bad_code, exc):
    with pytest.raises(exc):
        compiler.compile_code(bad_code)


valid_list = [
    """
@external
def foo(inp: Bytes[10]) -> Bytes[3]:
    return slice(inp, 2, 3)
    """,
    """
@external
def foo(inp: Bytes[10]) -> Bytes[4]:
    return slice(inp, 2, 3)
    """,
    """
@external
def foo() -> Bytes[10]:
    return slice(b"badmintonzzz", 1, 10)
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_slice_success(good_code):
    assert compiler.compile_code(good_code) is not None



from vyper.exceptions import CompilerPanic
from vyper.utils import OrderedSet
from vyper.venom.analysis.analysis import IRAnalysis
from vyper.venom.analysis.cfg import CFGAnalysis
from vyper.venom.basicblock import IRBasicBlock
from vyper.venom.function import IRFunction


class DominatorTreeAnalysis(IRAnalysis):
    """
    Dominator tree implementation. This class computes the dominator tree of a
    function and provides methods to query the tree. The tree is computed using
    the Lengauer-Tarjan algorithm.
    """

    fn: IRFunction
    entry_block: IRBasicBlock
    dfs_order: dict[IRBasicBlock, int]
    dfs_walk: list[IRBasicBlock]
    dominators: dict[IRBasicBlock, OrderedSet[IRBasicBlock]]
    immediate_dominators: dict[IRBasicBlock, IRBasicBlock]
    dominated: dict[IRBasicBlock, OrderedSet[IRBasicBlock]]
    dominator_frontiers: dict[IRBasicBlock, OrderedSet[IRBasicBlock]]

    def analyze(self):
        """
        Compute the dominator tree.
        """
        self.fn = self.function
        self.entry_block = self.fn.entry
        self.dfs_order = {}
        self.dfs_walk = []
        self.dominators = {}
        self.immediate_dominators = {}
        self.dominated = {}
        self.dominator_frontiers = {}

        self.analyses_cache.request_analysis(CFGAnalysis)

        self._compute_dfs(self.entry_block, OrderedSet())
        self._compute_dominators()
        self._compute_idoms()
        self._compute_df()

    def dominates(self, bb1, bb2):
        """
        Check if bb1 dominates bb2.
        """
        return bb2 in self.dominators[bb1]

    def immediate_dominator(self, bb):
        """
        Return the immediate dominator of a basic block.
        """
        return self.immediate_dominators.get(bb)

    def _compute_dominators(self):
        """
        Compute dominators
        """
        basic_blocks = list(self.dfs_order.keys())
        self.dominators = {bb: OrderedSet(basic_blocks) for bb in basic_blocks}
        self.dominators[self.entry_block] = OrderedSet([self.entry_block])
        changed = True
        count = len(basic_blocks) ** 2  # TODO: find a proper bound for this
        while changed:
            count -= 1
            if count < 0:
                raise CompilerPanic("Dominators computation failed to converge")
            changed = False
            for bb in basic_blocks:
                if bb == self.entry_block:
                    continue
                preds = bb.cfg_in
                if len(preds) == 0:
                    continue
                new_dominators = OrderedSet.intersection(*[self.dominators[pred] for pred in preds])
                new_dominators.add(bb)
                if new_dominators != self.dominators[bb]:
                    self.dominators[bb] = new_dominators
                    changed = True

    def _compute_idoms(self):
        """
        Compute immediate dominators
        """
        self.immediate_dominators = {bb: None for bb in self.dfs_order.keys()}
        self.immediate_dominators[self.entry_block] = self.entry_block
        for bb in self.dfs_walk:
            if bb == self.entry_block:
                continue
            doms = sorted(self.dominators[bb], key=lambda x: self.dfs_order[x])
            self.immediate_dominators[bb] = doms[1]

        self.dominated = {bb: OrderedSet() for bb in self.dfs_walk}
        for dom, target in self.immediate_dominators.items():
            self.dominated[target].add(dom)

    def _compute_df(self):
        """
        Compute dominance frontier
        """
        basic_blocks = self.dfs_walk
        self.dominator_frontiers = {bb: OrderedSet() for bb in basic_blocks}

        for bb in self.dfs_walk:
            if len(bb.cfg_in) > 1:
                for pred in bb.cfg_in:
                    runner = pred
                    while runner != self.immediate_dominators[bb]:
                        self.dominator_frontiers[runner].add(bb)
                        runner = self.immediate_dominators[runner]

    def dominance_frontier(self, basic_blocks: list[IRBasicBlock]) -> OrderedSet[IRBasicBlock]:
        """
        Compute dominance frontier of a set of basic blocks.
        """
        df = OrderedSet[IRBasicBlock]()
        for bb in basic_blocks:
            df.update(self.dominator_frontiers[bb])
        return df

    def _intersect(self, bb1, bb2):
        """
        Find the nearest common dominator of two basic blocks.
        """
        dfs_order = self.dfs_order
        while bb1 != bb2:
            while dfs_order[bb1] < dfs_order[bb2]:
                bb1 = self.immediate_dominators[bb1]
            while dfs_order[bb1] > dfs_order[bb2]:
                bb2 = self.immediate_dominators[bb2]
        return bb1

    def _compute_dfs(self, entry: IRBasicBlock, visited):
        """
        Depth-first search to compute the DFS order of the basic blocks. This
        is used to compute the dominator tree. The sequence of basic blocks in
        the DFS order is stored in `self.dfs_walk`. The DFS order of each basic
        block is stored in `self.dfs_order`.
        """
        visited.add(entry)

        for bb in entry.cfg_out:
            if bb not in visited:
                self._compute_dfs(bb, visited)

        self.dfs_walk.append(entry)
        self.dfs_order[entry] = len(self.dfs_walk)

    def as_graph(self) -> str:
        """
        Generate a graphviz representation of the dominator tree.
        """
        lines = ["digraph dominator_tree {"]
        for bb in self.fn.get_basic_blocks():
            if bb == self.entry_block:
                continue
            idom = self.immediate_dominator(bb)
            if idom is None:
                continue
            lines.append(f'    " {idom.label} " -> " {bb.label} "')
        lines.append("}")
        return "\n".join(lines)



import pytest

from vyper.cli.vyper_json import get_evm_version
from vyper.exceptions import JSONError


def test_unknown_evm():
    with pytest.raises(JSONError):
        get_evm_version({"settings": {"evmVersion": "foo"}})


@pytest.mark.parametrize(
    "evm_version_str",
    [
        "homestead",
        "tangerineWhistle",
        "spuriousDragon",
        "byzantium",
        "constantinople",
        "petersburg",
        "istanbul",
        "berlin",
    ],
)
def test_early_evm(evm_version_str):
    with pytest.raises(JSONError):
        get_evm_version({"settings": {"evmVersion": evm_version_str}})


@pytest.mark.parametrize("evm_version_str", ["london", "paris", "shanghai", "cancun"])
def test_valid_evm(evm_version_str):
    assert evm_version_str == get_evm_version({"settings": {"evmVersion": evm_version_str}})
